import numpy as np


def order_to_match(array, reference):
    """
    Given an array ``array``, return the index array needed to make it the same as ``reference``
    """
    return np.argsort(array)[np.argsort(np.argsort(reference))]



import os
import shutil
import tempfile

import numpy as np
from astropy.table import Table
from astropy import units as u


DATA = """
source_1 0.0 0.0 1 1 1 0.2 0.1 1.3 0.2 1.5 0.3
source_2 0.0 0.0 1 1 1 0.2 0.05 1.2 0.1 1.8 0.3
"""


def generate_random_models(models_dir, aperture_dependent=False):

    # Create fake SEDs
    np.random.seed(12345)
    os.mkdir(os.path.join(models_dir, 'seds'))

    names = []

    for i in range(5):

        from ..sed import SED

        sed = SED()

        sed.name = 'model_{0:04d}'.format(i)
        sed.distance = 1 * u.kpc
        sed.wav = np.logspace(-2., 3., 100) * u.micron
        sed.nu = sed.wav.to(u.Hz, equivalencies=u.spectral())

        if aperture_dependent:
            sed.apertures = np.logspace(1., 6., 10) * u.au
            sed.flux = np.cumsum(np.random.random((10, 100)), axis=0) * u.mJy
        else:
            sed.apertures = None
            sed.flux = (1 + np.random.random((1, 100))) * u.mJy

        sed.error = sed.flux * np.random.random(100) / 100.

        sed.write(os.path.join(models_dir, 'seds', sed.name + '_sed.fits'))

        names.append(sed.name)

    # Generate model conf file
    f = open(os.path.join(models_dir, 'models.conf'), 'w')
    f.write("name = test\n")
    f.write("length_subdir = 0\n")
    f.write("aperture_dependent = {0}\n".format('yes' if aperture_dependent else 'no'))
    f.write("logd_step = 0.02\n")
    f.close()

    # Generate model parameter file
    t = Table()
    t['MODEL_NAME'] = np.array(names, dtype='S30')
    t['par1'] = np.random.random(5)
    t['par2'] = np.random.random(5)
    t = t[[0, 4, 1, 3, 2]]
    t.write(os.path.join(models_dir, 'parameters.fits'))


class BasePipelineTest(object):

    def setup_class(self):

        self.tmpdir = tempfile.mkdtemp()
        generate_random_models(self.tmpdir, aperture_dependent=self.aperture_dependent)

        from ..extinction import Extinction

        self.extinction = Extinction()
        self.extinction.wav = np.logspace(-2., 3.) * u.micron
        self.extinction.chi = self.extinction.wav.value ** -2 * u.cm ** 2 / u.g

    def teardown_class(self):
        shutil.rmtree(self.tmpdir)

    def test_broadband(self, tmpdir):

        np.random.seed(12345)

        from ..filter import Filter

        f1 = Filter()
        f1.name = 'alice'
        f1.wavelength = 7. * u.micron
        f1.wav = np.linspace(5., 1., 100) * u.micron
        f1.nu = f1.wav.to(u.Hz, equivalencies=u.spectral())
        f1.r = np.random.random(100)
        f1.normalize()

        f2 = Filter()
        f2.name = 'bob'
        f2.wavelength = 12. * u.micron
        f2.wav = np.linspace(15., 10., 100) * u.micron
        f2.nu = f2.wav.to(u.Hz, equivalencies=u.spectral())
        f2.r = np.random.random(100)
        f2.normalize()

        f3 = Filter()
        f3.name = 'eve'
        f3.wavelength = 20. * u.micron
        f3.wav = np.linspace(25., 15., 100) * u.micron
        f3.nu = f3.wav.to(u.Hz, equivalencies=u.spectral())
        f3.r = np.random.random(100)
        f3.normalize()

        from ..convolve import convolve_model_dir

        convolve_model_dir(self.tmpdir, filters=[f1, f2, f3])

        from ..fit import fit

        data_file = tmpdir.join('data').strpath
        open(data_file, 'w').write(DATA.strip())

        output_file = tmpdir.join('output').strpath

        fit(data_file, ['bob', 'alice', 'eve'], [1., 3., 3.] * u.arcsec, self.tmpdir, output_file,
            extinction_law=self.extinction,
            distance_range=[1., 2.] * u.kpc,
            av_range=[0., 0.1],
            output_format=('F', 3.),
            output_convolved=False)

        self._postprocess(output_file, tmpdir)

    def test_monochromatic(self, tmpdir):

        from ..convolve import convolve_model_dir_monochromatic

        convolve_model_dir_monochromatic(self.tmpdir)

        from ..fit import fit

        data_file = tmpdir.join('data').strpath
        open(data_file, 'w').write(DATA.strip())

        output_file = tmpdir.join('output').strpath

        fit(data_file, ['MO001', 'MO002', 'MO020'], [1., 3., 3.] * u.arcsec, self.tmpdir, output_file,
            extinction_law=self.extinction,
            distance_range=[1., 2.] * u.kpc,
            av_range=[0., 0.1],
            output_format=('F', 3.),
            output_convolved=False)

        self._postprocess(output_file, tmpdir)

    def _postprocess(self, output_file, tmpdir):

        from ..plot import plot

        plots_dir = tmpdir.join('plots_sed_auto').strpath

        plot(output_file,
             plot_mode='A',
             output_dir=plots_dir,
             select_format=('F', 3.),
             format='png')

        plots_dir = tmpdir.join('plots_sed_manual').strpath

        plot(output_file,
             plot_mode='A',
             output_dir=plots_dir,
             select_format=('F', 3.),
             format='png',
             show_convolved=False, show_sed=True,
             x_mode='M', x_range=(0.1, 2000),
             y_mode='M', y_range=(1.e-14, 2e-8),
             plot_max=100)

        from ..plot_params_1d import plot_params_1d

        plots_dir = tmpdir.join('plots_1d').strpath

        plot_params_1d(output_file, 'par1',
                       log_x=True,
                       output_dir=plots_dir,
                       select_format=('F', 2.), format='png')

        from ..plot_params_2d import plot_params_2d

        plots_dir = tmpdir.join('plots_2d').strpath

        plot_params_2d(output_file, 'par1', 'par2',
                       log_x=True, log_y=True,
                       output_dir=plots_dir,
                       select_format=('F', 2.), format='png')

        from ..extract_parameters import extract_parameters

        output_prefix = tmpdir.join('extract_parameters').strpath

        extract_parameters(output_file, output_prefix)

        additional = {}
        additional['add'] = {}
        for i in range(5):
            additional['add']['model_{0:04d}'.format(i)] = np.random.random()

        from ..write_parameters import write_parameters

        output_ascii_file = tmpdir.join('write_parameters').strpath

        write_parameters(output_file, output_ascii_file)

        write_parameters(output_file, output_ascii_file, additional=additional)

        from ..write_parameter_ranges import write_parameter_ranges

        output_ascii_file = tmpdir.join('write_parameters_ranges').strpath

        write_parameter_ranges(output_file, output_ascii_file)

        write_parameter_ranges(output_file, output_ascii_file, additional=additional)

        from ..filter_output import filter_output

        filter_output(output_file, output_good='auto', output_bad='auto', cpd=3.)


class TestApertureIndependentPipeline(BasePipelineTest):
    aperture_dependent = False


class TestApertureDependentPipeline(BasePipelineTest):
    aperture_dependent = True



from matplotlib.ticker import LogFormatterMathtext


class LogFormatterMathtextAuto(LogFormatterMathtext):

    def __call__(self, x, pos=None):
        if x in [0.001, 0.01, 0.1]:
            return str(x)
        elif x in [1., 10., 100., 1000.]:
            return str(int(x))
        else:
            return LogFormatterMathtext.__call__(self, x, pos=pos)



from __future__ import print_function, division

import numpy as np

from .. import six


def is_numpy_array(variable):
    return type(variable) in [np.ndarray,
                              np.core.records.recarray,
                              np.ma.core.MaskedArray]


def validate_1d_array(name, value):

    if type(value) in [list, tuple]:
        value = np.array(value)
    if not is_numpy_array(value) or value.ndim != 1:
        raise ValueError(name + " should be a 1-D sequence")

    return value


class Source(object):

    def __init__(self):

        self.name = ""
        self.x = 0.
        self.y = 0.
        self.valid = None
        self.flux = None
        self.error = None

    def __getstate__(self):
        return {
            'name': self.name,
            'x': self.x,
            'y': self.y,
            'valid': self.valid,
            'flux': self.flux,
            'error': self.error
        }

    def __setstate__(self, d):
        self.__init__()
        self.name = d['name']
        self.x = d['x']
        self.y = d['y']
        self.valid = d['valid']
        self.flux = d['flux']
        self.error = d['error']

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is None:
            self._name = value
        elif isinstance(value, six.string_types):
            self._name = value
        else:
            raise TypeError("name should be a string")

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        if value is None:
            self._x = value
        elif np.isscalar(value) and np.isreal(value):
            self._x = value
        else:
            raise TypeError("x should be a scalar floating point value")

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        if value is None:
            self._y = value
        elif np.isscalar(value) and np.isreal(value):
            self._y = value
        else:
            raise TypeError("y should be a scalar floating point value")

    @property
    def valid(self):
        return self._valid

    @valid.setter
    def valid(self, value):
        if type(value) in [list, tuple]:
            value = np.array(value)
        if value is None:
            self._valid = value
        elif isinstance(value, np.ndarray) and value.ndim == 1:
            if self.n_wav is not None and len(value) != self.n_wav:
                raise ValueError("valid has incorrect length (expected {0} but found {1})".format(self.n_wav, len(value)))
            else:
                if np.any(value.astype(int) != value):
                    raise ValueError("valid values should be integers")
                elif np.any(value < 0) or np.any(value > 4):
                    raise ValueError("valid values should be in the range [0,4]")
                else:
                    self._valid = value
        else:
            raise TypeError("valid should be a 1-d sequence")

    @property
    def flux(self):
        return self._flux

    @flux.setter
    def flux(self, value):
        if type(value) in [list, tuple]:
            value = np.array(value)
        if value is None:
            self._flux = value
        elif isinstance(value, np.ndarray) and value.ndim == 1:
            if self.n_wav is not None and len(value) != self.n_wav:
                raise ValueError("flux has incorrect length (expected {0} but found {1})".format(self.n_wav, len(value)))
            else:
                self._flux = value
        else:
            raise TypeError("flux should be a 1-d sequence")

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, value):
        if type(value) in [list, tuple]:
            value = np.array(value)
        if value is None:
            self._error = value
        elif isinstance(value, np.ndarray) and value.ndim == 1:
            if self.n_wav is not None and len(value) != self.n_wav:
                raise ValueError("error has incorrect length (expected {0} but found {1})".format(self.n_wav, len(value)))
            else:
                self._error = value
        else:
            raise TypeError("error should be a 1-d sequence")

    @property
    def n_wav(self):
        if self.valid is not None:
            return len(self.valid)
        elif self.flux is not None:
            return len(self.flux)
        elif self.error is not None:
            return len(self.error)
        else:
            return None

    @property
    def n_data(self):
        return np.sum((self.valid == 1) | (self.valid == 4))

    def get_log_fluxes(self):

        # Initialize arrays
        log_flux = np.zeros(self.flux.shape, dtype=np.float64)
        log_error = np.zeros(self.error.shape, dtype=np.float64)
        weight = np.zeros(self.valid.shape, dtype=np.float64)

        # Fluxes
        r = self.valid == 1
        log_flux[r] = np.log10(self.flux[r]) - 0.5 * (self.error[r] / self.flux[r]) ** 2. / np.log(10.)
        log_error[r] = np.abs(self.error[r] / self.flux[r]) / np.log(10.)
        weight[r] = 1. / log_error[r] ** 2.

        # Lower and upper limits
        r = (self.valid == 2) | (self.valid == 3)
        log_flux[r] = np.log10(self.flux[r])
        log_error[r] = self.error[r]

        # Log10[Fluxes]
        r = self.valid == 4
        log_flux[r] = self.flux[r]
        log_error[r] = self.error[r]
        weight[r] = 1. / log_error[r] ** 2.

        # Ignored points
        r = self.valid == 9
        log_flux[r] = np.log10(self.flux[r]) - 0.5 * (self.error[r] / self.flux[r]) ** 2. / np.log(10.)
        log_error[r] = np.abs(self.error[r] / self.flux[r]) / np.log(10.)

        return weight, log_flux, log_error

    def __str__(self):

        weight, log_flux, log_error = self.get_log_fluxes()

        string = "Source name : %s\n" % self.name
        string += "x           : %9.5f\n" % self.x
        string += "y           : %9.5f\n" % self.y
        for j in range(self.n_wav):
            string += "F = %12.4e + / - %12.4e mJy (%1i)  Log[F] = %8.5f+ / -%8.5f\n" % \
                      (self.flux[j], self.error[j], self.valid[j], log_flux[j], log_error[j])

        return string

    @classmethod
    def from_ascii(cls, line):

        s = cls()

        cols = line.split()

        if len(cols) < 3:
            raise EOFError()

        s.name = cols[0]
        s.x = np.float64(cols[1])
        s.y = np.float64(cols[2])
        n_wav = np.int32((len(cols) - 3) / 3)
        s.valid = np.array(cols[3:3 + n_wav], dtype=int)
        flux_and_error = np.array(cols[3 + n_wav:], dtype=float)
        s.flux = flux_and_error[::2]
        s.error = flux_and_error[1::2]

        return s

    def to_ascii(self):
        line = ""
        line += "{0:30s} {1:9.5f} {2:9.5f} ".format(self.name, self.x, self.y)
        for v in self.valid:
            line += "{0:1d} ".format(v)
        for j in range(self.n_wav):
            line += "{0:11.3e} {1:11.3e} ".format(self.flux[j], self.error[j])
        return line

    @classmethod
    def from_dict(cls, source_dict):
        s = cls()
        s.name = source_dict['name']
        s.x = source_dict['x']
        s.y = source_dict['y']
        s.valid = source_dict['valid']
        s.flux = source_dict['flux']
        s.error = source_dict['error']
        return s

    def to_dict(self):
        return {
            'name': self.name,
            'x': self.x,
            'y': self.y,
            'valid': self.valid,
            'flux': self.flux,
            'error': self.error
        }

    def __eq__(self, other):
        return self.name == other.name and \
            self.x == other.x and \
            self.y == other.y and \
            np.all(self.valid == other.valid) and \
            np.all(self.flux == other.flux) and \
            np.all(self.error == other.error)



from __future__ import print_function, division

import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
from copy import deepcopy

from astropy.table import Table
from astropy.convolution import Tophat2DKernel

import numpy as np
from scipy.ndimage import convolve

import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties

from .fit_info import FitInfo
from .extinction import Extinction
from .models import load_parameter_table
from .utils import io
from .utils.formatter import LogFormatterMathtextAuto

KERNEL = Tophat2DKernel(5.5, x_size=11, y_size=11, mode='oversample').array
KERNEL /= KERNEL.max()  # normalize so maximum is 1

plt.rc('text', usetex=False)
plt.rc('axes', titlesize='small')
plt.rc('axes', labelsize='small')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('font', family='serif')
plt.rc('axes', linewidth=0.5)
plt.rc('patch', linewidth=0.5)

fp = FontProperties(size='small')


def get_axes(fig, label=None, zorder=None):
    vxmin, vxmax = 1.5, 4.5
    vymin, vymax = 1.0, 4.0
    width, height = fig.get_figwidth(), fig.get_figheight()
    rect = [vxmin / width, vymin / width, (vxmax - vxmin) / width, (vymax - vymin) / height]
    return fig.add_axes(rect, label=label, zorder=zorder)


def plot_params_2d(input_file, parameter_x, parameter_y, output_dir=None,
                   select_format=("N", 1), log_x=False, log_y=True,
                   label_x=None, label_y=None, additional={}, plot_name=True,
                   format='pdf'):
    """
    Make histogram plots of parameters

    Parameters
    ----------
    input_file : str
        File containing the fit information
    parameter_x : str
        The parameter to plot on the x-axis
    parameter_y : str
        The parameter to plot on the y-axis
    output_dir : str, optional
        If specified, plots are written to that directory
    select_format : tuple, optional
        Tuple specifying which fits should be plotted. See the documentation
        for a description of the tuple syntax.
    log_x : bool, optional
        Whether to plot the x-axis values in log space
    log_y : bool, optional
        Whether to plot the y-axis values in log space
    label_x : str, optional
        The x-axis label (if not specified, the parameter name is used)
    label_y : str, optional
        The y-axis label (if not specified, the parameter name is used)
    additional : dict, optional
        A dictionary specifying additional parameters not listed in the
        parameter list for the models. Each item of the dictionary should
        itself be a dictionary giving the values for each model (where the key
        is the model name).
    plot_name: bool, optional
        Whether to show the source name on the plot(s).
    format: str, optional
        The file format to use for the plot, if output_dir is specified.

    """

    npix = 1024

    if output_dir is None:
        raise ValueError("No output directory has been specified")

    # Create output directory
    io.create_dir(output_dir)

    # Open output file
    fin = open(input_file, 'rb')

    # Read in header of output file
    model_dir = pickle.load(fin)
    filters = pickle.load(fin)
    extinction = pickle.load(fin)

    # Read in table of parameters for model grid
    t = load_parameter_table(model_dir)

    # Sort alphabetically
    t['MODEL_NAME'] = np.char.strip(t['MODEL_NAME'])
    t.sort('MODEL_NAME')
    tpos = deepcopy(t)
    if log_x:
        tpos = tpos[tpos[parameter_x] > 0.]
    if log_y:
        tpos = tpos[tpos[parameter_y] > 0.]

    # Initialize figure
    fig = plt.figure()
    ax = get_axes(fig, label='main', zorder=0)

    # Find range of values
    xmin, xmax = tpos[parameter_x].min(), tpos[parameter_x].max()
    ymin, ymax = tpos[parameter_y].min(), tpos[parameter_y].max()

    # Compute histogram
    if log_x and log_y:
        gray_all, ex, ey = np.histogram2d(np.log10(tpos[parameter_x]),
                                          np.log10(tpos[parameter_y]), bins=npix,
                                          range=[[np.log10(xmin), np.log10(xmax)],
                                                 [np.log10(ymin), np.log10(ymax)]])
    elif log_x:
        gray_all, ex, ey = np.histogram2d(np.log10(tpos[parameter_x]),
                                          tpos[parameter_y], bins=npix,
                                          range=[[np.log10(xmin), np.log10(xmax)],
                                                 [ymin, ymax]])
    elif log_y:
        gray_all, ex, ey = np.histogram2d(tpos[parameter_x],
                                          np.log10(tpos[parameter_y]), bins=npix,
                                          range=[[xmin, xmax],
                                                 [np.log10(ymin), np.log10(ymax)]])
    else:
        gray_all, ex, ey = np.histogram2d(tpos[parameter_x],
                                          tpos[parameter_y], bins=npix,
                                          range=[[xmin, xmax],
                                                 [ymin, ymax]])

    gray_all = convolve(gray_all, KERNEL)
    gray_all = np.clip(gray_all, 0., 13.)

    # Grayscale showing all models. Since pcolormesh is very slow for PDF, we
    # create a 'ghost' axis which is already in log space.
    ax_log = get_axes(fig, label='log', zorder=-100)
    ax_log.axis('off')
    ax_log.imshow(gray_all.transpose(), cmap='binary', vmin=0, vmax=40., extent=[ex[0], ex[-1], ey[0], ey[-1]], aspect='auto', zorder=-100, origin='lower')

    ax.patch.set_facecolor('none')

    ax.set_xlabel(parameter_x if label_x is None else label_x)
    ax.set_ylabel(parameter_y if label_y is None else label_y)

    if log_x:
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(LogFormatterMathtextAuto())
    if log_y:
        ax.set_yscale('log')
        ax.yaxis.set_major_formatter(LogFormatterMathtextAuto())

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    if log_x:
        ax_log.set_xlim(np.log10(xmin), np.log10(xmax))
    else:
        ax_log.set_xlim(xmin, xmax)

    if log_y:
        ax_log.set_ylim(np.log10(ymin), np.log10(ymax))
    else:
        ax_log.set_ylim(ymin, ymax)

    ax.set_autoscale_on(False)
    ax_log.set_autoscale_on(False)

    pfits = None
    source_label = None

    while True:  # Loop over the fits

        # Read in next fit
        try:
            info = pickle.load(fin)
        except EOFError:
            break

        # Remove previous histogram
        if pfits is not None:
            pfits.remove()
        if source_label is not None:
            source_label.remove()

        # Filter fits
        info.keep(select_format[0], select_format[1])

        # Get filtered and sorted table of parameters
        tsorted = info.filter_table(t, additional=additional)

        pfits = ax.scatter(tsorted[parameter_x], tsorted[parameter_y], c='black', s=10)

        if plot_name:
            source_label = ax.text(0.5, 0.95, info.source.name,
                                   horizontalalignment='center',
                                   verticalalignment='center',
                                   transform=ax.transAxes,
                                   fontproperties=fp, zorder=200)
        # Save to file
        filename = "%s/%s.%s" % (output_dir, info.source.name, format)
        fig.savefig(filename, bbox_inches='tight', facecolor='none', dpi=300)

    # Close input and output files
    fin.close()

    plt.close(fig)



import pytest
import numpy as np

from .. import Source


def test_source_init():
    s = Source()


def test_source_name():
    s = Source()
    s.name = "Source Name"


@pytest.mark.parametrize('value', [1, 1., object(), np.array([1, 2, 3])])
def test_source_name_invalid(value):
    s = Source()
    with pytest.raises(TypeError) as exc:
        s.name = value
    assert exc.value.args[0] == 'name should be a string'


def test_source_position():
    s = Source()
    s.x = 1.
    s.y = 2.


@pytest.mark.parametrize('value', ['a', object(), np.array([1, 2, 3])])
def test_source_position_invalid(value):

    s = Source()

    with pytest.raises(TypeError) as exc:
        s.x = value
    assert exc.value.args[0] == "x should be a scalar floating point value"

    with pytest.raises(TypeError) as exc:
        s.y = value
    assert exc.value.args[0] == "y should be a scalar floating point value"


@pytest.mark.parametrize('value', [[1, 0, 1, 1, 2], (0, 1, 2, 3, 4), np.array([1., 2., 3.])])
def test_source_valid(value):
    s = Source()
    s.valid = value


@pytest.mark.parametrize('value', [1, 1., 'a', object()])
def test_source_valid_invalid(value):
    s = Source()
    with pytest.raises(TypeError) as exc:
        s.valid = value
    assert exc.value.args[0] == 'valid should be a 1-d sequence'


def test_source_valid_ints():
    s = Source()
    with pytest.raises(ValueError) as exc:
        s.valid = [1., 2.3, 4.]
    assert exc.value.args[0] == "valid values should be integers"


def test_source_valid_range():
    s = Source()
    with pytest.raises(ValueError) as exc:
        s.valid = [-1, 2, 3]
    assert exc.value.args[0] == "valid values should be in the range [0,4]"


@pytest.mark.parametrize(('attribute', 'value'), list(zip(['flux', 'error'], [[1., 2., 3., 4., 5], (1, 2, 3), np.array([1, 2, 3, 4])])))
def test_source_flux(attribute, value):
    s = Source()
    setattr(s, attribute, value)


@pytest.mark.parametrize(('attribute', 'value'), list(zip(['flux', 'error'], [1, 1., 'a', object()])))
def test_source_flux_invalid(attribute, value):
    s = Source()
    with pytest.raises(TypeError) as exc:
        setattr(s, attribute, value)
    assert exc.value.args[0] == '{0} should be a 1-d sequence'.format(attribute)


@pytest.mark.parametrize(('attribute'), ['valid', 'flux', 'error'])
def test_dimensions_match(attribute):
    s = Source()
    setattr(s, attribute, [1, 2, 3])
    for other in ['valid', 'flux', 'error']:
        if other != attribute:
            with pytest.raises(ValueError) as exc:
                setattr(s, other, [1, 2, 3, 4])
            assert exc.value.args[0] == "{0} has incorrect length (expected 3 but found 4)".format(other)


def test_dict():

    # Populate source object
    s = Source()
    s.name = "Source name"
    s.x = 1.
    s.y = 2.
    s.valid = [1, 0, 1]
    s.flux = [3., 4.4, 5.5]
    s.error = [0.1, 0.4, 0.3]
    d = s.to_dict()

    # Make sure the dictionary has the values we expect
    assert d['name'] == "Source name"
    assert d['x'] == 1.
    assert d['y'] == 2.
    assert np.all(d['valid'] == np.array([1, 0, 1]))
    assert np.all(d['flux'] == np.array([3., 4.4, 5.5]))
    assert np.all(d['error'] == np.array([0.1, 0.4, 0.3]))

    # Make sure that to_dict/from_dict round-trip
    assert Source.from_dict(d) == s


def test_get_log_fluxes():

    # Populate source object
    s = Source()
    s.name = "Source name"
    s.x = 1.
    s.y = 2.
    s.valid = [0, 1, 2, 3, 4]
    s.flux = [10., 10., 10., 10., 10.]
    s.error = [1., 1., 1., 1., 1.]

    weight, log_flux, log_error = s.get_log_fluxes()

    assert log_flux[0] == 0.
    assert log_flux[1] == 1. - 5.e-3 / np.log(10.)
    assert log_flux[2] == 1.
    assert log_flux[3] == 1.
    assert log_flux[4] == 10.

    assert log_error[0] == 0.
    assert log_error[1] == 0.1 / np.log(10.)
    assert log_error[2] == 1.
    assert log_error[3] == 1.
    assert log_error[4] == 1.

    assert weight[0] == 0.
    assert weight[1] == 1. / log_error[1] ** 2
    assert weight[2] == 0.
    assert weight[3] == 0.
    assert weight[4] == 1. / log_error[4] ** 2


def test_source_from_ascii():

    line = "a 1. 2. 0 1 1 2 0 1. 0.1 2. 0.2 3. 0.3 4. 0.4 5. 0.5"

    s = Source.from_ascii(line)

    assert s.name == "a"
    assert s.x == 1.
    assert s.y == 2.
    assert np.all(s.valid == np.array([0, 1, 1, 2, 0]))
    assert np.all(s.flux == np.array([1., 2., 3., 4., 5.]))
    assert np.all(s.error == np.array([0.1, 0.2, 0.3, 0.4, 0.5]))


def test_source_ascii_roundtrip():

    s = Source()
    s.name = "a"
    s.x = 1.
    s.y = 2.
    s.valid = [0, 1, 2, 3, 4]
    s.flux = [10., 20., 30., 40., 50.]
    s.error = [1., 0.4, 1.2, 1.3, 1.1]

    # Check that round-tripping works
    s2 = Source.from_ascii(s.to_ascii())
    assert s == s2



from __future__ import print_function, division

import os

import numpy as np
from astropy import log
from astropy.io import fits
from astropy.table import Table
from scipy.interpolate import interp1d
from astropy import units as u

from ..utils.validator import validate_array

UNIT_MAPPING = {}
UNIT_MAPPING['MICRONS'] = u.micron
UNIT_MAPPING['HZ'] = u.Hz
UNIT_MAPPING['MJY'] = u.mJy
UNIT_MAPPING['ergs/cm^2/s'] = u.erg / u.cm ** 2 / u.s


def parse_unit_safe(unit_string):
    if unit_string in UNIT_MAPPING:
        return UNIT_MAPPING[unit_string]
    else:
        return u.Unit(unit_string, parse_strict=False)


def assert_allclose_quantity(q1, q2):
    np.testing.assert_allclose(q1.value, q2.to(q1.unit).value)


class SED(object):

    def __init__(self):

        self.name = None
        self.distance = None
        self.wav = None
        self.nu = None
        self.apertures = None
        self.flux = None
        self.error = None

    def __eq__(self, other):

        try:
            assert self.name == other.name
            assert_allclose_quantity(self.distance, other.distance)
            assert_allclose_quantity(self.distance, other.distance)
            assert_allclose_quantity(self.wav, other.wav)
            assert_allclose_quantity(self.nu, other.nu)
            assert_allclose_quantity(self.apertures, other.apertures)
            assert_allclose_quantity(self.flux, other.flux)
            assert_allclose_quantity(self.error, other.error)
        except AssertionError:
            return False
        else:
            return True

    def copy(self):
        from copy import deepcopy
        return deepcopy(self)

    def scale_to_distance(self, distance):
        """
        Returns the SED scaled to distance `distance`

        Parameters
        ----------
        distance : float
            The distance in cm

        Returns
        -------
        sed : SED
            The SED, scaled to the new distance
        """
        sed = self.copy()
        sed.distance = distance * u.cm
        sed.flux = sed.flux * (self.distance.to(u.cm) / sed.distance) ** 2
        sed.error = sed.error * (self.distance.to(u.cm) / sed.distance) ** 2
        return sed

    def scale_to_av(self, av, law):
        sed = self.copy()
        sed.flux = sed.flux * 10. ** (av * law(sed.wav))
        sed.error = sed.error * 10. ** (av * law(sed.wav))
        return sed

    @property
    def wav(self):
        """
        The wavelengths at which the SED is defined
        """
        return self._wav

    @wav.setter
    def wav(self, value):
        if value is None:
            self._wav = None
        else:
            self._wav = validate_array('wav', value, domain='positive', ndim=1,
                                       shape=None if self.nu is None else (len(self.nu),),
                                       physical_type='length')

    @property
    def nu(self):
        """
        The frequencies at which the SED is defined
        """
        return self._nu

    @nu.setter
    def nu(self, value):
        if value is None:
            self._nu = None
        else:
            self._nu = validate_array('nu', value, domain='positive', ndim=1,
                                      shape=None if self.wav is None else (len(self.wav),),
                                      physical_type='frequency')

    @property
    def apertures(self):
        """
        The apertures at which the SED is defined
        """
        return self._apertures

    @apertures.setter
    def apertures(self, value):
        if value is None:
            self._apertures = None
        else:
            self._apertures = validate_array('apertures', value, domain='positive',
                                             ndim=1, physical_type='length')

    @property
    def flux(self):
        """
        The SED fluxes
        """
        return self._flux

    @flux.setter
    def flux(self, value):
        if value is None:
            self._flux = value
        else:
            self._flux = validate_array('flux', value, ndim=2,
                                        shape=(self.n_ap, self.n_wav),
                                        physical_type=('power', 'flux', 'spectral flux density'))

    @property
    def error(self):
        """
        The convolved flux errors
        """
        return self._error

    @error.setter
    def error(self, value):
        if value is None:
            self._error = value
        else:
            self._error = validate_array('error', value, ndim=2,
                                         shape=(self.n_ap, self.n_wav),
                                         physical_type=('power', 'flux', 'spectral flux density'))

    @property
    def n_ap(self):
        if self.apertures is None:
            return 1
        else:
            return len(self.apertures)

    @property
    def n_wav(self):
        if self.wav is None:
            return None
        else:
            return len(self.wav)

    @classmethod
    def read(cls, filename, unit_wav=u.micron, unit_freq=u.Hz,
             unit_flux=u.erg / u.cm ** 2 / u.s, order='nu'):
        """
        Read an SED from a FITS file.

        Parameters
        ----------
        filename: str
            The name of the file to read the SED from.
        unit_wav: `~astropy.units.Unit`, optional
            The units to convert the wavelengths to.
        unit_freq: `~astropy.units.Unit`, optional
            The units to convert the frequency to.
        unit_flux: `~astropy.units.Unit`, optional
            The units to convert the flux to.
        order: str, optional
            Whether to sort the SED by increasing wavelength (`wav`) or
            frequency ('nu').
        """

        # Instantiate SED class
        sed = cls()

        # Assume that the filename may be missing the .gz extension
        if not os.path.exists(filename) and os.path.exists(filename + '.gz'):
            filename += ".gz"

        # Open FILE file
        hdulist = fits.open(filename, memmap=False)

        # Extract model name
        sed.name = hdulist[0].header['MODEL']

        # Check if distance is specified in header, otherwise assume 1kpc
        if 'DISTANCE' in hdulist[0].header:
            sed.distance = hdulist[0].header['DISTANCE'] * u.cm
        else:
            log.debug("No distance found in SED file, assuming 1kpc")
            sed.distance = 1. * u.kpc

        # Extract SED values
        wav = hdulist[1].data.field('WAVELENGTH')
        nu = hdulist[1].data.field('FREQUENCY')
        ap = hdulist[2].data.field('APERTURE')
        flux = hdulist[3].data.field('TOTAL_FLUX')
        error = hdulist[3].data.field('TOTAL_FLUX_ERR')

        # Set units

        wav = wav * parse_unit_safe(hdulist[1].columns[0].unit)
        nu = nu * parse_unit_safe(hdulist[1].columns[1].unit)
        ap = ap * parse_unit_safe(hdulist[2].columns[0].unit)
        flux = flux * parse_unit_safe(hdulist[3].columns[0].unit)
        error = error * parse_unit_safe(hdulist[3].columns[1].unit)

        sed.apertures = ap

        if flux.unit != error.unit:
            raise ValueError("flux and flux error unit should match")

        # Convert wavelength and frequencies to requested units
        sed.wav = wav.to(unit_wav)
        sed.nu = nu.to(unit_freq)

        # Convert flux and error

        if not unit_flux.is_equivalent(flux.unit):

            # Convert to ergs / cm^2 / s

            if flux.unit.is_equivalent(u.erg / u.s):
                flux = flux / sed.distance ** 2
                error = error / sed.distance ** 2
            elif flux.unit.is_equivalent(u.Jy):
                flux = flux * nu
                error = error * nu
            elif not flux.unit.is_equivalent(u.erg / u.cm ** 2 / u.s):
                raise Exception("Don't know how to convert {0} to ergs/cm^2/s" % (flux.unit))

            # Convert to requested unit

            if unit_flux.is_equivalent(u.erg / u.s):
                flux = flux * sed.distance ** 2
                error = error * sed.distance ** 2
            elif unit_flux.is_equivalent(u.Jy):
                flux = flux / nu
                error = error / nu
            elif not flux.unit.is_equivalent(u.erg / u.cm ** 2 / u.s):
                raise Exception("Don't know how to convert %s to %s" % (curr_unit_flux, unit_flux))

        sed.flux = flux.to(unit_flux)
        sed.error = error.to(unit_flux)

        # Sort SED

        if order not in ('nu', 'wav'):
            raise ValueError('order should be nu or wav')

        if (order == 'nu' and sed.nu[0] > sed.nu[-1]) or \
           (order == 'wav' and sed.wav[0] > sed.wav[-1]):
            sed.wav = sed.wav[::-1]
            sed.nu = sed.nu[::-1]
            sed.flux = sed.flux[..., ::-1]
            sed.error = sed.error[..., ::-1]

        return sed

    def write(self, filename):
        """
        Write an SED to a FITS file.

        Parameters
        ----------
        filename: str
            The name of the file to write the SED to.
        """

        # Create first HDU with meta-data
        hdu0 = fits.PrimaryHDU()

        if self.name is None:
            raise ValueError("Model name is not set")
        else:
            hdu0.header['MODEL'] = self.name

        if self.distance is None:
            raise ValueError("Model distance is not set")
        else:
            hdu0.header['DISTANCE'] = self.distance.to(u.cm).value

        hdu0.header['NAP'] = self.n_ap
        hdu0.header['NWAV'] = self.n_wav

        # Create wavelength table
        twav = Table()
        if self.wav is None:
            raise ValueError("Wavelengths are not set")
        else:
            twav['WAVELENGTH'] = self.wav
        if self.nu is None:
            raise ValueError("Frequencies are not set")
        else:
            twav['FREQUENCY'] = self.nu
        twav.sort('FREQUENCY')
        hdu1 = fits.BinTableHDU(np.array(twav))
        hdu1.columns[0].unit = self.wav.unit.to_string(format='fits')
        hdu1.columns[1].unit = self.nu.unit.to_string(format='fits')

        # Create aperture table
        tap = Table()
        if self.apertures is None:
            tap['APERTURE'] = [1.e-30]
        else:
            tap['APERTURE'] = self.apertures
        hdu2 = fits.BinTableHDU(np.array(tap))
        if self.apertures is None:
            hdu2.columns[0].unit = 'cm'
        else:
            hdu2.columns[0].unit = self.apertures.unit.to_string(format='fits')

        # Create flux table
        tflux = Table()
        if self.flux is None:
            raise ValueError("Fluxes are not set")
        else:
            tflux['TOTAL_FLUX'] = self.flux
        if self.error is None:
            raise ValueError("Wavelengths are not set")
        else:
            tflux['TOTAL_FLUX_ERR'] = self.error
        hdu3 = fits.BinTableHDU(np.array(tflux))
        hdu3.columns[0].unit = self.flux.unit.to_string(format='fits')
        hdu3.columns[1].unit = self.error.unit.to_string(format='fits')

        # Create overall FITS file
        hdulist = fits.HDUList([hdu0, hdu1, hdu2, hdu3])
        hdulist.writeto(filename)

    def interpolate(self, apertures):
        """
        Interpolate the SED to different apertures
        """

        # If there is only one aperture, we can't interpolate, we can only repeat
        if self.n_ap == 1:
            return np.repeat(self.flux[0, :], len(apertures)).reshape(self.n_wav, len(apertures))

        # Create interpolating function
        flux_interp = interp1d(self.apertures, self.flux.swapaxes(0, 1))

        # If any apertures are larger than the defined max, reset to max
        apertures[apertures > self.apertures.max()] = self.apertures.max()

        # If any apertures are smaller than the defined min, raise Exception
        if np.any(apertures < self.apertures.min()):
            raise Exception("Aperture(s) requested too small")

        return flux_interp(apertures)

    def interpolate_variable(self, wavelengths, apertures):
        """
        Interpolate the SED to a variable aperture as a function of
        wavelength. This method should be called with an interpolating
        function for aperture as a function of wavelength, in log10 space.
        """

        if self.n_ap == 1:
            return self.flux[0, :]

        sed_apertures = self.apertures.to(u.au).value
        sed_wav = self.wav.to(u.micron).value

        # If any apertures are larger than the defined max, reset to max
        apertures[apertures > sed_apertures.max()] = sed_apertures.max() * 0.999

        # If any apertures are smaller than the defined min, raise Exception
        if np.any(apertures < sed_apertures.min()):
            raise Exception("Aperture(s) requested too small")

        # Find wavelength order
        order = np.argsort(wavelengths)

        # Interpolate apertures vs wavelength
        log10_ap_interp = interp1d(np.log10(wavelengths[order]), np.log10(apertures[order]), bounds_error=False, fill_value=np.nan)

        # Create interpolating function
        flux_interp = interp1d(sed_apertures, self.flux.swapaxes(0, 1))

        # Interpolate the apertures
        apertures = 10. ** log10_ap_interp(np.log10(sed_wav))

        # Extrapolate on either side
        apertures[np.log10(sed_wav) < log10_ap_interp.x[0]] = 10. ** log10_ap_interp.y[0]
        apertures[np.log10(sed_wav) > log10_ap_interp.x[-1]] = 10. ** log10_ap_interp.y[-1]

        # Interpolate and return only diagonal elements
        return flux_interp(apertures).diagonal()



from __future__ import print_function, division

import os
import glob

import numpy as np

from astropy.io import fits
from astropy.logger import log
from astropy import units as u
from astropy.utils.console import ProgressBar

from ..convolved_fluxes import ConvolvedFluxes
from ..sed import SED
from ..models import load_parameter_table
from .. import six


def convolve_model_dir(model_dir, filters, overwrite=False):
    """
    Convolve all the model SEDs in a model directory

    Parameters
    ----------
    model_dir: str
        The path to the model directory
    filters: list
        A list of Filter objects to use for the convolution
    overwrite: bool, optional
        Whether to overwrite the output files
    """

    for f in filters:
        if f.name is None:
            raise Exception("filter name needs to be set")
        if f.wavelength is None:
            raise Exception("filter central wavelength needs to be set")

    # Create 'convolved' sub-directory if needed
    if not os.path.exists(model_dir + '/convolved'):
        os.mkdir(model_dir + '/convolved')

    # Find all SED files to convolve
    sed_files = (glob.glob(model_dir + '/seds/*.fits.gz') +
                 glob.glob(model_dir + '/seds/*/*.fits.gz') +
                 glob.glob(model_dir + '/seds/*.fits') +
                 glob.glob(model_dir + '/seds/*/*.fits'))

    par_table = load_parameter_table(model_dir)

    if len(sed_files) == 0:
        raise Exception("No SEDs found in %s" % model_dir)
    else:
        log.info("{0} SEDs found in {1}".format(len(sed_files), model_dir))

    # Find out apertures
    first_sed = SED.read(sed_files[0])
    n_ap = first_sed.n_ap
    apertures = first_sed.apertures

    # Set up convolved fluxes
    fluxes = [ConvolvedFluxes(model_names=np.zeros(len(sed_files), dtype='U30' if six.PY3 else 'S30'), apertures=apertures, initialize_arrays=True) for i in range(len(filters))]

    # Set up list of binned filters
    binned_filters = []
    binned_nu = None

    # Loop over SEDs

    b = ProgressBar(len(sed_files))

    for im, sed_file in enumerate(sed_files):

        log.debug('Convolving {0}'.format(os.path.basename(sed_file)))

        # Read in SED
        s = SED.read(sed_file, unit_freq=u.Hz, unit_flux=u.mJy, order='nu')

        # Check if filters need to be re-binned
        try:
            assert binned_nu is not None
            np.testing.assert_array_almost_equal_nulp(s.nu.value, binned_nu.value, 100)
        except AssertionError:
            log.info('Rebinning filters')
            binned_filters = [f.rebin(s.nu) for f in filters]
            binned_nu = s.nu

        b.update()

        # Convolve
        for i, f in enumerate(binned_filters):

            fluxes[i].wavelength = f.wavelength
            fluxes[i].apertures = apertures
            fluxes[i].model_names[im] = s.name

            if n_ap == 1:
                fluxes[i].flux[im] = np.sum(s.flux * f.r)
                fluxes[i].error[im] = np.sqrt(np.sum((s.error * f.r) ** 2))
            else:
                fluxes[i].flux[im, :] = np.sum(s.flux * f.r, axis=1)
                fluxes[i].error[im] = np.sqrt(np.sum((s.error * f.r) ** 2, axis=1))

    for i, f in enumerate(binned_filters):
        fluxes[i].sort_to_match(par_table['MODEL_NAME'])
        fluxes[i].write(model_dir + '/convolved/' + f.name + '.fits',
                        overwrite=overwrite)



from __future__ import print_function, division

import os
import sys
import shutil

from ..six.moves import input


def create_dir(dir_name):
    delete_dir(dir_name)
    os.system("mkdir " + dir_name)


def delete_dir(dir_name):

    if os.path.exists(dir_name):
        reply = input("Delete directory " + dir_name + "? [y/[n]] ")
        if reply == 'y':
            shutil.rmtree(dir_name)
        else:
            print("Aborting...")
            sys.exit()
        print("")


def delete_file(file_name):

    if os.path.exists(file_name):
        reply = input("Delete file " + file_name + "? [y/[n]] ")
        if reply == 'y':
            os.remove(file_name)
        else:
            print("Aborting...")
            sys.exit()
        print("")



import os
import pytest

from ..parfile import read

ROOT = os.path.dirname(__file__)

@pytest.mark.parametrize('format', ('par', 'conf'))
def test_parfile(format):
    p = read(os.path.join(ROOT, 'data', 'test.' + format), format)
    assert p['a'] == 'hello, world!'
    assert p['b'] == 1
    assert p['c'] == 2.3
    assert p['d'] == '1 + 2'



from __future__ import print_function, division

import numpy as np
np.seterr(all='ignore')

from scipy.interpolate import interp1d

from astropy.logger import log
from astropy import units as u
from astropy.utils.misc import isiterable

from ..utils.validator import validate_scalar, validate_array

# TODO: get rid of use of interp1d


def is_numpy_array(variable):
    return issubclass(variable.__class__, (np.ndarray,
                                           np.core.records.recarray,
                                           np.ma.core.MaskedArray))


class ConvolvedFluxes(object):

    def __init__(self, wavelength=None, model_names=None, apertures=None,
                 flux=None, error=None,
                 initialize_arrays=False, initialize_units=u.mJy):

        self.model_names = model_names
        self.apertures = apertures
        self.wavelength = wavelength

        if initialize_arrays:

            if model_names is None:
                raise ValueError("model_names is required when using initialize_arrays=True")

            if apertures is None:
                raise ValueError("apertures is required when using initialize_arrays=True")

            if flux is None:
                self.flux = np.zeros((self.n_models, self.n_ap)) * initialize_units
            else:
                self.flux = flux

            if error is None:
                self.error = np.zeros((self.n_models, self.n_ap)) * initialize_units
            else:
                self.error = error

        else:

            self.flux = flux
            self.error = error

    @property
    def wavelength(self):
        """
        The central or characteristic wavelength of the filter
        """
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        if value is None:
            self._wavelength = None
        else:
            self._wavelength = validate_scalar('wavelength', value, domain='strictly-positive', physical_type='length')

    @property
    def model_names(self):
        """
        The names of the models
        """
        return self._model_names

    @model_names.setter
    def model_names(self, value):
        if value is None:
            self._model_names = value
        else:
            self._model_names = validate_array('model_names', value, ndim=1)

    @property
    def apertures(self):
        """
        The apertures at which the SED is defined
        """
        return self._apertures

    @apertures.setter
    def apertures(self, value):
        if value is None:
            self._apertures = None
        else:
            self._apertures = validate_array('apertures', value, domain='positive', ndim=1, physical_type='length')

    @property
    def flux(self):
        """
        The SED fluxes
        """
        return self._flux

    @flux.setter
    def flux(self, value):
        if value is None:
            self._flux = value
        else:

            if self.model_names is None:
                raise ValueError("model_names has not been set")

            self._flux = validate_array('flux', value, ndim=2,
                                        shape=(self.n_models, self.n_ap),
                                        physical_type=('power', 'flux', 'spectral flux density'))

    @property
    def error(self):
        """
        The convolved flux errors
        """
        return self._error

    @error.setter
    def error(self, value):
        if value is None:
            self._error = value
        else:

            if self.model_names is None:
                raise ValueError("model_names has not been set")

            self._error = validate_array('error', value, ndim=2,
                                         shape=(self.n_models, self.n_ap),
                                         physical_type=('power', 'flux', 'spectral flux density'))

    @property
    def n_models(self):
        if self.model_names is None:
            return None
        else:
            return self.model_names.shape[0]

    @property
    def n_ap(self):
        if self.apertures is None:
            return 1
        else:
            return len(self.apertures)

    def __eq__(self, other):
        return self.wavelength == other.wavelength \
            and np.all(self.model_names == other.model_names) \
            and np.all(self.apertures == other.apertures) \
            and np.all(self.flux == other.flux) \
            and np.all(self.error == other.error)

    def sort_to_match(self, requested_model_names):
        """
        Sort the models following ``requested_model_names``
        """

        order = np.arange(self.n_models)

        # Make sure there are no spaces at the end of model names
        requested_model_names = np.char.strip(requested_model_names)

        from ..utils.misc import order_to_match
        order = order_to_match(self.model_names, requested_model_names)

        # Double check that the sorting will work
        if not np.all(self.model_names[order] == requested_model_names):
            raise Exception("Sorting failed")

        self.model_names = self.model_names[order]
        self.flux = self.flux[order, :]
        self.error = self.error[order, :]

    @classmethod
    def read(cls, filename):
        """
        Read convolved flux from a FITS file

        Parameters
        ----------
        filename : str
            The name of the FITS file to read the convolved fluxes from
        """

        from astropy.io import fits
        from astropy.table import Table

        conv = cls()

        # Open the convolved flux FITS file
        convolved = fits.open(filename)

        keywords = convolved[0].header

        # Try and read in the wavelength of the filter
        if 'FILTWAV' in keywords:
            conv.wavelength = keywords['FILTWAV'] * u.micron
        else:
            conv.wavelength = None

        # Read in apertures, if present
        try:
            ta = Table.read(convolved['APERTURES'])
            if ta['APERTURE'].unit is None:
                ta['APERTURE'].unit = u.au
            conv.apertures = ta['APERTURE'].data * ta['APERTURE'].unit
        except KeyError:
            pass

        # Create shortcuts to table
        tc = Table.read(convolved['CONVOLVED FLUXES'])

        # Read in model names
        conv.model_names = tc['MODEL_NAME']

        # Read in flux and flux errors

        if tc['TOTAL_FLUX'].unit is None:
            tc['TOTAL_FLUX'].unit = u.mJy

        if tc['TOTAL_FLUX_ERR'].unit is None:
            tc['TOTAL_FLUX_ERR'].unit = u.mJy

        if tc['TOTAL_FLUX'].ndim == 1 and conv.n_ap == 1:
            conv.flux = tc['TOTAL_FLUX'].data.reshape(tc['TOTAL_FLUX'].shape[0], 1) * tc['TOTAL_FLUX'].unit
        else:
            conv.flux = tc['TOTAL_FLUX'].data * tc['TOTAL_FLUX'].unit

        if tc['TOTAL_FLUX_ERR'].ndim == 1 and conv.n_ap == 1:
            conv.error = tc['TOTAL_FLUX_ERR'].data.reshape(tc['TOTAL_FLUX_ERR'].shape[0], 1) * tc['TOTAL_FLUX_ERR'].unit
        else:
            conv.error = tc['TOTAL_FLUX_ERR'].data * tc['TOTAL_FLUX_ERR'].unit

        # Read in 99% cumulative and 50% surface brightness radii
        try:
            if tc['RADIUS_SIGMA_50'].unit is None:
                tc['RADIUS_SIGMA_50'].unit = u.au
            if tc['RADIUS_CUMUL_99'].unit is None:
                tc['RADIUS_CUMUL_99'].unit = u.au
            conv.radius_sigma_50 = tc['RADIUS_SIGMA_50'] * tc['RADIUS_SIGMA_50'].unit
            conv.radius_cumul_99 = tc['RADIUS_CUMUL_99'] * tc['RADIUS_CUMUL_99'].unit
        except KeyError:
            pass

        return conv

    def write(self, filename, overwrite=False):
        """
        Write convolved flux to a FITS file.

        Parameters
        ----------
        filename: str
            The name of the file to output the convolved fluxes to.
        overwrite: bool, optional
            Whether to overwrite the output file
        """

        from astropy.io import fits
        from astropy.table import Table, Column

        tc = Table()
        tc['MODEL_NAME'] = self.model_names.astype('S30')
        tc['TOTAL_FLUX'] = self.flux
        tc['TOTAL_FLUX_ERR'] = self.error

        if self.apertures is not None:
            radius_sigma_50 = self.find_radius_sigma(0.50)
            tc['RADIUS_SIGMA_50'] = radius_sigma_50
            radius_cumul_99 = self.find_radius_cumul(0.99)
            tc['RADIUS_CUMUL_99'] = radius_cumul_99

        if self.apertures is not None:
            ta = Table()
            ta['APERTURE'] = self.apertures

        # Primary HDU (for metadata)
        hdu0 = fits.PrimaryHDU()
        if self.wavelength is not None:
            hdu0.header['FILTWAV'] = self.wavelength.to(u.micron).value
        hdu0.header['NMODELS'] = self.n_models
        hdu0.header['NAP'] = self.n_ap

        # Convolved fluxes
        hdu1 = fits.BinTableHDU(np.array(tc), name='CONVOLVED FLUXES')
        hdu1.columns[1].unit = self.flux.unit.to_string(format='fits')
        hdu1.columns[2].unit = self.error.unit.to_string(format='fits')
        if self.apertures is not None:
            hdu1.columns[3].unit = radius_sigma_50.unit.to_string(format='fits')
            hdu1.columns[4].unit = radius_cumul_99.unit.to_string(format='fits')

        # Apertures
        if self.apertures is not None:
            hdu2 = fits.BinTableHDU(np.array(ta), name='APERTURES')
            hdu2.columns[0].unit = self.apertures.unit.to_string(format='fits')
        else:
            hdu2 = None

        hdulist = fits.HDUList([hdu0, hdu1] if hdu2 is None else [hdu0, hdu1, hdu2])
        hdulist.writeto(filename, clobber=overwrite)

    def interpolate(self, apertures):
        """
        Interpolate the flux to the apertures specified.

        Parameters
        ----------
        apertures : `astropy.units.Quantity` instance
            The apertures to interpolate to
        """

        # Initalize new ConvolvedFluxes object to return
        c = ConvolvedFluxes()

        # Set wavelength
        c.wavelength = self.wavelength

        # Save requested apertures
        c.apertures = apertures[:]

        # Transfer model names
        c.model_names = self.model_names

        # Interpolate to requested apertures
        if self.n_ap > 1:

            # If any apertures are larger than the defined max, reset to max
            if np.any(c.apertures > self.apertures.max()):
                apertures[c.apertures > self.apertures.max()] = self.apertures.max()

            # If any apertures are smaller than the defined min, raise error
            if np.any(c.apertures < self.apertures.min()):
                raise Exception("Aperture(s) requested too small")

            # Note that we have to be careful here because interp1d will drop
            # the units, so we need to make sure the new apertures are in the
            # same units as the current ones for the interpolation, and we need
            # to add the flux unit back.

            flux_interp = interp1d(self.apertures, self.flux)
            c.flux = flux_interp(c.apertures.to(self.apertures.unit)) * self.flux.unit

            # The following is not strictly correct - errors from interpolation is not interpolation of errors
            error_interp = interp1d(self.apertures, self.error)
            c.error = error_interp(c.apertures.to(self.apertures.unit)) * self.error.unit

        else:

            c.flux = np.repeat(self.flux, len(c.apertures)).reshape(c.n_models, len(c.apertures))
            c.error = np.repeat(self.error, len(c.apertures)).reshape(c.n_models, len(c.apertures))

        try:
            # TODO: is this actually correct?!
            c.radius_sigma_50 = self.radius_sigma_50[:]
            c.radius_cumul_99 = self.radius_cumul_99[:]
        except AttributeError:
            pass

        return c

    def find_radius_cumul(self, fraction):
        """
        Find for each model the radius containing a fraction of the flux.

        Parameters
        ----------
        fraction: float
            The fraction to use when determining the radius
        """

        log.info("Calculating radii containing %g%s of the flux" % (fraction * 100., '%'))

        radius = np.zeros(self.n_models, dtype=self.flux.dtype) * u.au

        if self.apertures is None:

            return radius

        else:

            required = fraction * self.flux[:, -1]

            # Linear interpolation - need to loop over apertures for vectorization
            for ia in range(len(self.apertures) - 1):
                calc = (required >= self.flux[:, ia]) & (required < self.flux[:, ia + 1])
                radius[calc] = (required[calc] - self.flux[calc, ia]) / \
                               (self.flux[calc, ia + 1] - self.flux[calc, ia]) * \
                               (self.apertures[ia + 1] - self.apertures[ia]) + \
                    self.apertures[ia]

            calc = (required < self.flux[:, 0])
            radius[calc] = self.apertures[0]

            calc = (required >= self.flux[:, -1])
            radius[calc] = self.apertures[-1]

            return radius

    def find_radius_sigma(self, fraction):
        """
        Find for each model a fractional surface brightness radius

        This is the outermost radius where the surface brightness is larger
        than a fraction of the peak surface brightness.

        Parameters
        ----------
        fraction: float
            The fraction to use when determining the radius
        """

        log.info("Calculating %g%s peak surface brightness radii" % (fraction * 100., '%'))

        sigma = np.zeros(self.flux.shape, dtype=self.flux.dtype)
        sigma[:, 0] = self.flux[:, 0] / self.apertures[0] ** 2
        sigma[:, 1:] = (self.flux[:, 1:] - self.flux[:, :-1]) / \
                       (self.apertures[1:] ** 2 - self.apertures[:-1] ** 2)

        maximum = np.max(sigma, axis=1)

        radius = np.zeros(self.n_models, dtype=self.flux.dtype) * u.au

        # Linear interpolation - need to loop over apertures backwards for vectorization
        for ia in range(len(self.apertures) - 2, -1, -1):
            calc = (sigma[:, ia] > fraction * maximum) & (radius == 0.)
            radius[calc] = (sigma[calc, ia] - fraction * maximum[calc]) / \
                           (sigma[calc, ia] - sigma[calc, ia + 1]) * \
                           (self.apertures[ia + 1] - self.apertures[ia]) + \
                self.apertures[ia]

        calc = sigma[:, -1] > fraction * maximum
        radius[calc] = self.apertures[-1]

        return radius



from __future__ import print_function, division

# Still to implement:
# - Performance monitoring
# - Remove resolved models
# - Optional FITS input/output
# - Output convolved fluxes

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np
from astropy import units as u

from . import timer

from .models import Models
from .fit_info import FitInfo
from .source import Source
from .utils import io
from .utils.validator import validate_scalar, validate_array
from . import six


def fit(data, filter_names, apertures, model_dir, output, n_data_min=3,
        extinction_law=None, av_range=None, distance_range=None,
        output_format=('F', 6.), output_convolved=False,
        remove_resolved=False):
    """
    Fit a set of sources with models

    Parameters
    ----------
    data : str
        Filename of the file containing the data, one source per line (see
        documentation for a description of the required format).
    filter_names : tuple or list
        List of filter names (given as individual strings) for which the data
        is defined. The filter names should be the name of the files in the
        ``convolved`` directory for the models, without the extensions. This is
        typically ``2J``, ``I1``, ``M1``, etc.
    apertures : :class`~astropy.units.quantity.Quantity` array instance
        The aperture radii that the data is specified in (as an angle). The
        fluxes may not be measured from aperture photometry, but this is meant
        to give an indication of the sizescale of the emission, and can be used
        to reject models that would have been clearly resolved at the distance
        specified.
    models_dir : str
        Name of the directory containing the models to use.
    output : str
        Name of the file to output the fit information to (in binary format).
    extinction_law : :class:`~sedfitter.extinction.Extinction` instance
        The extinction law to use.
    av_range : tuple
        Minimum and maximum Av to allow in the fitting.
    distance_range : :class`~astropy.units.quantity.Quantity` array instance
        Minimum and maximum distance to allow in the fitting in units of length.
    n_data_min : int, optional
        The minimum number of points a source needs to be fit.
    output_format : tuple, optional
        Tuple specifying which fits should be output. See the documentation
        for a description of the tuple syntax.
    output_convolved : bool, optional
        Whether to output the convolved fluxes (necessary if the convolved
        model fluxes are needed for the SED plot).
    remove_resolved : bool, optional
        If set, then models larger than the aperture are removed. See
        Robitaille et al. (2007) for a discussion of this criterion.
    """

    validate_array('apertures', apertures, domain='positive', ndim=1, physical_type='angle')
    validate_array('distance_range', distance_range, domain='positive', ndim=1, shape=(2,), physical_type='length')

    print(" ------------------------------------------------------------")
    print("  => Fitting parameters")
    print(" ------------------------------------------------------------")
    print("")
    print("   Minimum A_V      : %9.3f mag" % av_range[0])
    print("   Maximum A_V      : %9.3f mag" % av_range[1])
    print("   Minimum distance : %9.3f %s" % (distance_range[0].value, distance_range.unit))
    print("   Maximum distance : %9.3f %s" % (distance_range[1].value, distance_range.unit))
    print("")
    print(" ------------------------------------------------------------")
    print("  => Output parameters")
    print(" ------------------------------------------------------------")
    print("")
    print("   File   : %s" % output)
    print("   Format : %s" % output_format[0])
    print("   Number : %g" % output_format[1])
    print("")
    print(" ------------------------------------------------------------")
    print("  => Data format parameters")
    print(" ------------------------------------------------------------")
    print("")
    print("   Number of filters :  %i" % len(filter_names))
    print("")

    # Open datafile
    if isinstance(data, six.string_types):
        data_file = open(data, 'r')
    else:
        data_file = data

    if len(apertures) != len(filter_names):
        raise ValueError("length of apertures list should match length of filter names list")

    # Construct filters dictionary
    filters = []
    for i in range(len(apertures)):
        filters.append({'aperture_arcsec': apertures[i].to(u.arcsec).value, 'name': filter_names[i]})

    # Read in models
    models = Models.read(model_dir, filters, distance_range=distance_range, remove_resolved=remove_resolved)

    # Add wavelength to filters
    for i, f in enumerate(filters):
        f['wav'] = models.wavelengths[i]

    print('')
    print('     Filter    Wavelength    Aperture (")   ')
    print('    ----------------------------------------')
    for f in filters:
        print('       %5s   %9.2f  %9.2f        ' % (f['name'], f['wav'].to(u.micron).value, f['aperture_arcsec']))
    print('')

    # Set Av law
    av_law = extinction_law.get_av(models.wavelengths)

    # Set scale model - make this a scalar
    sc_law = -2. * np.ones(av_law.shape)

    # Cycle through sources

    io.delete_file(output)

    fout = open(output, 'wb')
    pickle.dump(model_dir, fout, 2)
    pickle.dump(filters, fout, 2)
    pickle.dump(extinction_law, fout, 2)

    # NOTE _ CAN USE PROTOCOL 2 IN SOURCE FOR COORINDATES

    s = Source()

    t = timer.Timer()

    while True:

        try:
            s = Source.from_ascii(data_file.readline())
        except EOFError:
            break

        if s.n_data >= n_data_min:

            info = FitInfo(source=s)
            info.av, info.sc, info.chi2, info.model_name, model_fluxes = models.fit(s, av_law, sc_law, av_range[0], av_range[1])

            if output_convolved:
                info.model_fluxes = model_fluxes

            info.sort()
            info.keep(output_format[0], output_format[1])

            pickle.dump(info, fout, 2)

            t.display()

    t.display(force=True)

    fout.close()



from __future__ import print_function, division

import os
import glob

import numpy as np
from astropy.io import fits
from astropy.logger import log

from ..source import Source
from .. import six


def list2data(file_in, file_out, models_dir=None, source_name='source'):
    """
    Convert a list-formatted data file into the standard fitting format

    In some cases it can be convenient to specify an SED in a list-based
    format with one source per file, one wavelength per line. This script
    converts the former into the latter.

    Parameters
    ----------
    file_in: str or file
        The input file.
    file_out: str or file
        The output file.
    models_dir: str, optional
        The models directory that will be used to fit the source. This can be
        useful to check that the desired filters exist, and to find out which
        monochromatic filters can be used.
    source_name: str, optional
        The source name to use. Defaults to 'source'.

    Returns
    -------
    filter_names: list of str
        The list of filters for the datafile (can be passed to fit())
    apertures: list of float
        The list of apertures for the datafile (can be passed to fit())
    """

    # Load datafile
    datafile = np.loadtxt(file_in, dtype=[('flux_type', int), ('name', 'S30'),
                                          ('valid', int),
                                          ('flux', float), ('err', float),
                                          ('aperture', float)])

    # Read in monochromatic wavelengths if available
    if models_dir is not None:

        log.info('Reading in monochromatic wavelengths')

        filter_files = glob.glob(os.path.join(models_dir, 'convolved/MO*'))

        mono_wav = np.zeros(len(filter_files), dtype=float)
        mono_filter = np.zeros(len(filter_files), dtype='S30')

        for i, filter_file in enumerate(filter_files):

            log.debug('Reading {}'.format(filter_file))

            mono_wav[i] = fits.getheader(filter_file)['FILTWAV']
            mono_filter[i] = os.path.basename(filter_file).replace('.fits', '').replace('.gz', '')

    elif np.any(datafile['flux_type'] == 2):

        raise Exception("list contains monochromatic fluxes, so need " +
                        "models_dir to be specified")

    # Create source
    s = Source()
    s.name = source_name
    s.x = 0.
    s.y = 0.
    s.n_wav = len(datafile)
    s.valid = datafile['valid']
    s.flux = datafile['flux']
    s.error = datafile['err']
    if isinstance(file_out, six.string_types):
        s.write_ascii(open(file_out, 'wb'))
    else:
        s.write_ascii(file_out)

    # Determine filter names
    for i in range(len(datafile)):
        if datafile['flux_type'][i] == 2:
            wav = float(datafile['name'][i])
            j = np.argmin(np.abs(mono_wav - wav))
            datafile['name'][i] = mono_filter[j]

    return datafile['name'].tolist(), datafile['aperture'].tolist()



from __future__ import print_function, division

import numpy as np

from .interpolate import interp1d_fast


def integrate_subset(x, y, xmin, xmax):
    """
    Perform trapezium integration of a set of points (x,y) between bounds xmin
    and xmax. The interpolation between the points is done in linear space, so
    this is designed for functions that are piecewise linear in linear space.
    """

    # Swap arrays if necessary
    if x[-1] < x[0]:
        x = x[::-1]
        y = y[::-1]

    # Swap limits if necessary
    if xmin > xmax:
        xmin, xmax = xmax, xmin
    elif xmin == xmax:
        return 0.

    # Find the subset of points to use and the value of the function at the
    # end-points of the integration

    if xmin == x[0]:
        i1 = 1
        ymin = y[0]
    else:
        i1 = np.searchsorted(x, xmin)
        ymin = interp1d_fast(x[i1 - 1:i1 + 1], y[i1 - 1:i1 + 1], xmin)

    if xmax == x[-1]:
        i2 = -2
        ymax = y[-1]
    else:
        i2 = np.searchsorted(x, xmax)
        ymax = interp1d_fast(x[i2 - 1:i2 + 1], y[i2 - 1:i2 + 1], xmax)

    # Construct sub-arrays of the relevant data
    x = np.hstack([xmin, x[i1:i2], xmax])
    y = np.hstack([ymin, y[i1:i2], ymax])

    # Call function to integrate the whole subset
    return integrate(x, y)


def integrate(x, y):
    """
    Perform trapezium integration of a set of points (x,y). The interpolation
    between the points is done in linear space, so this is designed for
    functions that are piecewise linear in linear space.
    """

    # Fix NaN values
    y[np.isnan(y)] = 0.

    # Find the integral of all the chunks
    integrals = 0.5 * (x[1:] - x[:-1]) * (y[1:] + y[:-1])

    # Sum them all up
    integral = np.sum(integrals)

    # Check if the integral is NaN or infinity
    if np.isnan(integral) or np.isinf(integral):
        raise Exception("Integral is NaN or Inf")

    return integral



from __future__ import print_function, division

import numpy as np


def linear_regression(data, weights, pattern1, pattern2):

    c1 = np.sum(data * pattern1 * weights, axis=1)
    c2 = np.sum(data * pattern2 * weights, axis=1)
    m11 = np.sum(pattern1 * pattern1 * weights)
    m12 = np.sum(pattern1 * pattern2 * weights)
    m22 = np.sum(pattern2 * pattern2 * weights)

    inv_det = 1. / (m11 * m22 - m12 * m12)

    p1 = (m22 * c1 - m12 * c2) * inv_det
    p2 = (m11 * c2 - m12 * c1) * inv_det

    return p1, p2


def optimal_scaling(data, weights, pattern1):
    return np.sum(data * pattern1 * weights, axis=data.ndim - 1) / \
        np.sum(pattern1 * pattern1 * weights)


def chi_squared(valid, data, error, weight, model):
    """
    valid has dimension (n_wav, )
    data has dimension (n_wav, n_models)
    error has dimension(n_wav)
    weight has dimension(n_wav)
    model has dimension (n_wav, n_models)
    """

    # Calculate the 'default' chi^2 and handle special cases after
    chi2_array = (data - model) ** 2 * weight

    # Force chi^2 to zero for valid == 0
    if chi2_array.ndim == 2:
        chi2_array[:, valid == 0] = 0.
    elif chi2_array.ndim == 3:
        chi2_array[:, :, valid == 0] = 0.
    else:
        raise Exception("Chi^2 array has unexpected number of dimensions: %i" % chi2_array.ndim)

    # Reset lower limits where model < data
    if chi2_array.ndim == 2:
        for j in np.where(valid == 2)[0]:
            reset = model[:, j] < data[:, j]
            chi2_array[:, j][reset] = -2. * np.log(1. - error[j])
    else:
        for j in np.where(valid == 2)[0]:
            reset = model[:, :, j] < data[:, :, j]
            chi2_array[:, :, j][reset] = -2. * np.log(1. - error[j])

    # Reset upper limits where model > data
    if chi2_array.ndim == 2:
        for j in np.where(valid == 3)[0]:
            reset = model[:, j] > data[:, j]
            chi2_array[:, j][reset] = -2. * np.log(1. - error[j])
    else:
        for j in np.where(valid == 3)[0]:
            reset = model[:, :, j] > data[:, :, j]
            chi2_array[:, :, j][reset] = -2. * np.log(1. - error[j])

    # Check that there are no infinities
    chi2_array[np.isinf(chi2_array)] = 1.e30

    return np.sum(chi2_array, axis=chi2_array.ndim - 1)



from __future__ import print_function, division

# Still to implement:
# - Axis numbers - done
# - Interpolation of SEDs (stype) - done - needs testing
# - Use number/format to know how many to plot (can use info.keep) - done - needs testing

# Low priority
# - Grayscale SEDs
# - Color of lines
# - Option to plot/not plot SEDs (pseds)
# - Plot convolved fluxes (pconv)
# - Overplot stellar photosphere (patmo)
# - Color caption for multi-aperture plot
# - Ignore device (or translate to matplotlib)
# - Esthetic parameters (warn that they are being ignored for now)

# Optimization:
# - Compute interpolating functions as little as possible

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np
from astropy import units as u

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.font_manager import FontProperties
import matplotlib.ticker as ticker

from .extinction import Extinction
from .fit_info import FitInfo
from .sed import SED
from .utils import io
from .utils import parfile
from .utils.formatter import LogFormatterMathtextAuto

KPC = 3.086e21

plt.rc('text', usetex=True)
plt.rc('axes', titlesize='small')
plt.rc('axes', labelsize='small')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('font', family='serif')
plt.rc('axes', linewidth=0.5)
plt.rc('patch', linewidth=0.5)

fp = FontProperties(size='x-small')

color = {}

color['gray'] = '0.75'
color['black'] = '0.00'

color['full'] = []
color['full'].append((0.65, 0.00, 0.00))
color['full'].append((0.20, 0.30, 0.80))
color['full'].append((1.00, 0.30, 0.30))
color['full'].append((1.00, 0.30, 0.60))
color['full'].append((0.30, 0.80, 0.30))
color['full'].append((0.50, 0.10, 0.80))
color['full'].append((0.20, 0.60, 0.80))
color['full'].append((1.00, 0.00, 0.00))
color['full'].append((0.50, 0.25, 0.00))
color['full'].append((0.90, 0.90, 0.00))
color['full'].append((0.00, 0.50, 0.00))

color['faded'] = []
color['faded'].append((1.00, 0.70, 0.70))
color['faded'].append((0.70, 0.70, 0.80))
color['faded'].append((1.00, 0.80, 0.70))
color['faded'].append((1.00, 0.75, 1.00))
color['faded'].append((0.70, 0.80, 0.70))
color['faded'].append((0.75, 0.60, 0.80))
color['faded'].append((0.70, 0.75, 0.80))
color['faded'].append((1.00, 0.70, 0.80))
color['faded'].append((0.90, 0.80, 0.70))
color['faded'].append((0.90, 0.90, 0.70))
color['faded'].append((0.50, 0.90, 0.50))


def tex_friendly(string):
    return string
    # return string.replace('_', '\_').replace('%', '\%')


def plot_source_info(ax, i, info, plot_name, plot_info):

    labels = []

    if plot_name:
        labels.append(tex_friendly(info.source.name))

    if plot_info:
        labels.append("Model: %s" % tex_friendly(info.model_name[i]))
        if i == 0:
            labels.append("Best fit")
        else:
            labels.append("Fit: %i" % (i + 1))
        labels.append("$\chi^2$ = %10.3f    A$_{\\rm V}$ = %5.1f    Scale = %5.2f" % (info.chi2[i], info.av[i], info.sc[i]))

    pos = 0.95
    for label in labels:
        ax.text(0.5, pos, label, horizontalalignment='center',
                verticalalignment='center',
                transform=ax.transAxes,
                fontproperties=fp)
        pos -= 0.06

    return ax


def plot_source_data(ax, source, filters, size=20, capsize=3):

    weight, log_flux, log_error = source.get_log_fluxes()

    wav = np.array([f['wav'].to(u.micron).value for f in filters])
    plot_wav = wav
    plot_flux = log_flux - 26. + np.log10(3.e8 / (wav * 1.e-6))
    plot_flux_up = 10. ** (plot_flux + log_error)
    plot_flux_down = 10. ** (plot_flux - log_error)
    plot_flux = 10. ** plot_flux
    plot_error = np.vstack([plot_flux - plot_flux_down, plot_flux_up - plot_flux])

    for j in range(source.n_wav):

        if source.valid[j] in [1, 4]:
            ax.scatter(plot_wav[j], plot_flux[j], marker='o', edgecolor='black', facecolor='black', zorder=100, s=size)
            ax.errorbar(plot_wav[j], plot_flux[j], yerr=plot_error[:, j:j + 1], color='black', zorder=100, capsize=capsize)
        elif source.valid[j] == 2:
            ax.scatter(plot_wav[j], plot_flux[j], marker='^', edgecolor='black', facecolor='black', zorder=100, s=size)
        elif source.valid[j] == 3:
            ax.scatter(plot_wav[j], plot_flux[j], marker='v', edgecolor='black', facecolor='black', zorder=100, s=size)
        elif source.valid[j] == 9:
            ax.scatter(plot_wav[j], plot_flux[j], marker='o', edgecolor='black', facecolor='none', zorder=100, s=size)

    return ax


def set_view_limits(ax, wav, source, x_mode, y_mode, x_range, y_range):

    if x_mode == 'A':
        xmin = np.min(wav[source.valid != 0]) * 10. ** (-x_range[0])
        xmax = np.max(wav[source.valid != 0]) * 10. ** (+x_range[1])
    else:
        xmin = x_range[0]
        xmax = x_range[1]

    if y_mode == 'A':
        weight, log_flux, log_error = source.get_log_fluxes()
        plot_flux = 10. ** (log_flux - 26. + np.log10(3.e8 / (wav * 1.e-6)))
        ymin = np.min(plot_flux[source.valid != 0]) * 10. ** (-y_range[0])
        ymax = np.max(plot_flux[source.valid != 0]) * 10. ** (+y_range[1])
    else:
        ymin = y_range[0]
        ymax = y_range[1]

    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim((xmin, xmax))
    ax.set_ylim((ymin, ymax))
    ax.xaxis.set_major_formatter(LogFormatterMathtextAuto())

    return ax


def get_axes(fig):
    vxmin, vxmax = 1.5, 5.5
    vymin, vymax = 1.0, 4.0
    width, height = fig.get_figwidth(), fig.get_figheight()
    rect = [vxmin / width, vymin / width, (vxmax - vxmin) / width, (vymax - vymin) / height]
    return fig.add_axes(rect)


def plot(input_file, output_dir=None, select_format=("N", 1), plot_max=None,
         plot_mode="A", sed_type="interp", show_sed=True, show_convolved=False,
         x_mode='A', y_mode='A', x_range=(1., 1.), y_range=(1., 2.),
         plot_name=True, plot_info=True, format='pdf'):
    """
    Make SED plots

    Parameters
    ----------
    input_file : str
        File containing the fit information
    output_dir : str, optional
        If specified, plots are written to that directory
    select_format : tuple, optional
        Tuple specifying which fits should be plotted. See the documentation
        for a description of the tuple syntax.
    plot_max : int, optional
        Maximum number of fits to plot
    plot_mode : str, optional
        Whether to plot all fits in a single plot ('A') or one fit per plot
        ('I')
    sed_type : str, optional
        Which SED should be shown:
            * `largest`: show the SED for the largest aperture specified.
            * `smallest`: show the SED for the smallest aperture specified.
            * `largest+smallest`: show the SEDs for the largest and smallest
              apertures specified.
            * `all`: show the SEDs for all apertures specified.
            * `interp`: interpolate the SEDs to the correct aperture at each
              wavelength (and interpolated apertures in between), so that a
              single composite SED is shown.
    show_sed : bool, optional
        Show the SEDs
    show_convolved : bool, optional
        Show convolved model fluxes
    x_mode : str, optional
        Whether to automatically select the wavelength range ('A'), or whether
        to use manually set values ('M').
    x_range : tuple, optional
        If x_mode is set to 'M', this is the range of wavelengths to show. If
        x_mode is set to 'A', this is the marging to add around the wavelength
        range (in dex).
    y_mode : str, optional
        Whether to automatically select the flux range ('A'), or whether to
        use manually set values ('M').
    y_range : tuple, optional
        If y_mode is set to 'M', this is the range of fluxes to show. If
        y_mode is set to 'A', this is the marging to add around the flux
        range (in dex).
    plot_name : bool, optional
        Whether to show the source name on the plot(s).
    plot_info : bool, optional
        Whether to show the fit information on the plot(s).
    format : str, optional
        The file format to use for the plot, if output_dir is specified.
    """

    if output_dir:
        io.create_dir(output_dir)
    else:
        figures = {}

    fin = open(input_file, 'rb')

    model_dir = pickle.load(fin)
    filters = pickle.load(fin)
    extinction = pickle.load(fin)

    wav = np.array([f['wav'].to(u.micron).value for f in filters])
    ap = np.array([f['aperture_arcsec'] for f in filters])

    unique_ap = np.unique(ap)
    unique_ap.sort()

    # Read in model parameters
    modpar = parfile.read("%s/models.conf" % model_dir, 'conf')

    while True:

        # Read in next fit
        try:
            info = pickle.load(fin)
        except EOFError:
            break

        # Filter fits
        info.keep(select_format[0], select_format[1])

        if plot_max:
            info.keep('N', plot_max)

        if show_convolved and info.model_fluxes is None:
            raise Exception("Cannot plot convolved fluxes as these are not included in the input file")

        for i in range(info.n_fits - 1, -1, -1):

            if (plot_mode == 'A' and i == info.n_fits - 1) or plot_mode == 'I':
                fig = plt.figure()
                ax = get_axes(fig)

            # Initalize lines and colors list
            if (plot_mode == 'A' and i == info.n_fits - 1) or plot_mode == 'I':
                lines = []
                colors = []
                if show_convolved:
                    conv = []

            if (plot_mode == 'A' and i == 0) or plot_mode == 'I':
                if sed_type in ['interp', 'largest']:
                    color_type = 'black'
                else:
                    color_type = 'full'
            else:
                if sed_type in ['interp', 'largest']:
                    color_type = 'gray'
                else:
                    color_type = 'faded'

            if modpar['length_subdir'] == 0:
                s = SED.read(model_dir + '/seds/' + info.model_name[i] + '_sed.fits')
            else:
                s = SED.read(model_dir + '/seds/%s/%s_sed.fits' % (info.model_name[i][:modpar['length_subdir']], info.model_name[i]))

            s = s.scale_to_distance(10. ** info.sc[i] * KPC)
            s = s.scale_to_av(info.av[i], extinction.get_av)

            if sed_type == 'interp':
                apertures = ap * 10. ** info.sc[i] * 1000.
                flux = s.interpolate_variable(wav, apertures)
            elif sed_type == 'largest':
                apertures = np.array([ap.max()]) * 10. ** info.sc[i] * 1000.
                flux = s.interpolate(apertures)
            elif sed_type == 'largest+smallest':
                apertures = np.array([ap.min(), ap.max()]) * 10. ** info.sc[i] * 1000.
                flux = s.interpolate(apertures)
            elif sed_type == 'all':
                apertures = unique_ap * 10. ** info.sc[i] * 1000.
                flux = s.interpolate(apertures)

            if flux.ndim > 1:
                for j in range(flux.shape[1]):
                    lines.append(np.column_stack([s.wav, flux[:, j]]))
                    colors.append(color[color_type][j])
            else:
                lines.append(np.column_stack([s.wav, flux]))
                colors.append(color[color_type])

            if show_convolved:
                conv.append(10. ** (info.model_fluxes[i, :] - 26. + np.log10(3.e8 / (wav * 1.e-6))))

            if (plot_mode == 'A' and i == 0) or plot_mode == 'I':

                if show_sed:
                    ax.add_collection(LineCollection(lines, colors=colors))

                if show_convolved:
                    for j in range(len(conv)):
                        ax.plot(wav, conv[j], color=colors[j], linestyle='solid', marker='o', markerfacecolor='none', markeredgecolor=colors[j])

                ax = plot_source_data(ax, info.source, filters)

                if plot_mode == 'A':
                    ax = plot_source_info(ax, 0, info, plot_name, plot_info)
                else:
                    ax = plot_source_info(ax, i, info, plot_name, plot_info)

                ax.set_xlabel('$\lambda$ ($\mu$m)')
                ax.set_ylabel('$\lambda$F$_\lambda$ (ergs/cm$^2$/s)')

                ax = set_view_limits(ax, wav, info.source, x_mode, y_mode, x_range, y_range)

                if output_dir:
                    if plot_mode == 'A':
                        filename = "%s/%s.%s" % (output_dir, info.source.name, format)
                    else:
                        filename = "%s/%s_%05i.%s" % (output_dir, info.source.name, i + 1, format)
                    fig.savefig(filename, bbox_inches='tight')
                    plt.close(fig)
                else:
                    figures[info.source.name] = {'source': info.source, 'filters': filters, 'lines': LineCollection(lines, colors=colors)}

    if not output_dir:
        return figures



from __future__ import print_function, division

import numpy as np


class check_bounds(object):
    """
    Decorator to add standard interpolation bounds checking.

    This decorator incurs an overhead of 30-50 percent on the runtime of
    the interpolation routine.
    """

    def __init__(self, f):
        self.f = f

    def __call__(self, x, y, xval, bounds_error=True, fill_value=np.nan):
        if np.isscalar(xval):  # xval is a scalar
            if xval < x[0] or xval > x[-1]:  # the value is out of bounds
                if bounds_error:
                    raise Exception("x value is out of interpolation bounds")
                else:
                    return np.nan
            else:  # the value is in the bounds
                return self.f(x, y, xval)
        else:  # xval is an array
            inside = (xval >= x[0]) & (xval <= x[-1])
            outside = ~inside
            if np.any(outside):  # some values are out of bounds
                if bounds_error:
                    raise Exception("x values are out of interpolation bounds")
                else:
                    if np.any(inside):
                        yval = np.zeros(xval.shape)
                        yval[inside] = self.f(x, y, xval[inside])
                        yval[outside] = fill_value
                        return yval
                    else:
                        return np.repeat(fill_value, xval.shape)
            else:  # all values are in the bounds
                return self.f(x, y, xval)


@check_bounds
def interp1d_fast(x, y, xval):
    """On-the-fly linear interpolator"""
    if len(x) != len(y):
        raise Exception("x and y should have the same length")
    ipos = np.searchsorted(x, xval)
    return (xval - x[ipos - 1]) \
        / (x[ipos] - x[ipos - 1]) \
        * (y[ipos] - y[ipos - 1]) \
        + y[ipos - 1]



from __future__ import print_function, division


def read(filename, format):
    """
    Read a parameter file and return a dictionary
    """

    parameters = {}

    for line in open(filename):
        if '=' in line and line[0] != "#" and line.strip() != "":
            cols = line.split('=')
            if format == 'par':
                value, key = cols[0].strip(), cols[1].strip()
            elif format == 'conf':
                key, value = cols[0].strip(), cols[1].strip()
            else:
                raise Exception("Format should be par or conf")
            try:
                value = int(value)
            except:
                try:
                    value = float(value)
                except:
                    if value.lower() in ['y', 'yes']:
                        value = True
                    elif value.lower() in ['n', 'no']:
                        value = False
            parameters[key] = value

    return parameters



from __future__ import print_function, division

import os

import numpy as np
from astropy import units as u

from ..utils.integrate import integrate_subset, integrate
from ..utils.validator import validate_array, validate_scalar

c = 299792458.


class Filter(object):

    def __init__(self):
        self.name = None
        self.wavelength = None
        self.wav = None
        self.nu = None
        self.r = None

    @classmethod
    def read(cls, filename):
        """
        Read a filter from a file

        Parameters
        ----------
        filename: str
            The name of the file containing the filter
        """

        self = cls()

        # Read in central wavelength
        self.wavelength = float(open(filename, 'r').readline().split('=')[1]) * u.micron

        # Read in spectral response curve
        self.wav = np.loadtxt(filename, usecols=[0], dtype=float) * u.micron
        self.r = np.loadtxt(filename, usecols=[1], dtype=float)

        # Compute frequency
        self.nu = self.wav.to(u.Hz, equivalencies=u.spectral())

        # Set name
        if self.name is None:
            self.name = os.path.basename(filename).split('.')[0]

        return self

    def normalize(self):
        """
        Normalize so the integral over nu is 1
        """
        self.r /= integrate(self.nu.to(u.Hz).value, self.r)

    def rebin(self, nu_new):
        """
        Re-bin the filter onto a new frequency grid

        Parameters
        ----------
        nu_new: np.ndarray
            The new frequency grid

        Returns
        -------
        filter: Filter
            The binned filter
        """

        # Define new filter
        f = Filter()
        f.name = self.name
        f.wavelength = self.wavelength
        f.nu = nu_new
        f.wav = f.nu.to(u.micron, equivalencies=u.spectral())

        self_nu_hz = self.nu.to(u.Hz).value
        nu_new_hz = f.nu.to(u.Hz).value

        # Compute re-binned transmission

        f.r = np.zeros(nu_new.shape)

        for i in range(len(f.r)):

            if i == 0:
                nu1 = nu_new_hz[0]
            else:
                nu1 = 0.5 * (nu_new_hz[i - 1] + nu_new_hz[i])

            if i == len(nu_new_hz) - 1:
                nu2 = nu_new_hz[-1]
            else:
                nu2 = 0.5 * (nu_new_hz[i] + nu_new_hz[i + 1])

            nu1 = min(max(nu1, self_nu_hz[0]), self_nu_hz[-1])
            nu2 = min(max(nu2, self_nu_hz[0]), self_nu_hz[-1])

            if nu2 != nu1:
                f.r[i] = integrate_subset(self_nu_hz, self.r, nu1, nu2)

        return f

    @property
    def wavelength(self):
        """
        The central or characteristic wavelength of the filter
        """
        return self._wavelength

    @wavelength.setter
    def wavelength(self, value):
        if value is None:
            self._wavelength = None
        else:
            self._wavelength = validate_scalar('wavelength', value, domain='strictly-positive', physical_type='length')

    @property
    def wav(self):
        """
        The wavelengths at which the filter is defined
        """
        return self._wav

    @wav.setter
    def wav(self, value):
        if value is None:
            self._wav = None
        else:
            self._wav = validate_array('wav', value, domain='strictly-positive', ndim=1,
                                       shape=None if self.nu is None else (len(self.nu),),
                                       physical_type='length')

    @property
    def nu(self):
        """
        The frequencies at which the filter is defined
        """
        return self._nu

    @nu.setter
    def nu(self, value):
        if value is None:
            self._nu = None
        else:
            self._nu = validate_array('nu', value, domain='strictly-positive', ndim=1,
                                      shape=None if self.wav is None else (len(self.wav),),
                                      physical_type='frequency')

    @property
    def r(self):
        """
        The filter response
        """
        return self._r

    @r.setter
    def r(self, value):
        if value is None:
            self._r = None
        else:
            self._r = validate_array('r', value, domain='positive', ndim=1,
                                     shape=None if (self.wav or self.nu) is None else (len(self.wav or self.nu),))



import os

import pytest
import numpy as np

from astropy import units as u

from .. import SED

ROOT = os.path.dirname(__file__)


def test_init():
    SED()


def test_wav():
    s = SED()
    s.wav = [1,2,3] * u.m


@pytest.mark.parametrize('value', [1. * u.m, [[1,2,3],[4,5,6]] * u.m])
def test_wav_invalid_shape(value):
    s = SED()
    with pytest.raises(TypeError) as exc:
        s.wav = value
    assert exc.value.args[0] == 'wav should be a 1-d sequence'


@pytest.mark.parametrize('value', ['string', object(), 1., [1,2,3]])
def test_wav_invalid_type(value):
    s = SED()
    with pytest.raises(TypeError) as exc:
        s.wav = value
    assert exc.value.args[0] == 'wav should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.Jy])
def test_wav_invalid_unit(unit):
    s = SED()
    with pytest.raises(TypeError) as exc:
        s.wav = [1., 2., 3.] * unit
    assert exc.value.args[0] == 'wav should be given in units of length'


# def test_aperture_none():
#     s = SED()
#     s.apertures = None
#
#
# def test_aperture_list():
#     s = SED()
#     s.apertures = [1., 2., 3.]
#
#
# @pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
# def test_aperture_invalid_type(value):
#
#     s = SED()
#
#     with pytest.raises(TypeError) as exc:
#         s.apertures = value
#     assert exc.value.args[0] == 'apertures should be a 1-d sequence'
#
#
# def test_model_name_list():
#     s = SED()
#     s.model_names = ['a', 'b', 'c']
#
#
# @pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
# def test_model_name_invalid_type(value):
#
#     s = SED()
#
#     with pytest.raises(TypeError) as exc:
#         s.model_names = value
#     assert exc.value.args[0] == 'model_names should be a 1-d sequence'
#
#
# def test_model_name_unset():
#
#     s = SED()
#
#     with pytest.raises(ValueError) as exc:
#         s.flux = [1., 2., 3.]
#     assert exc.value.args[0] == "model_names has not been set"
#
#     with pytest.raises(ValueError) as exc:
#         s.error = [1., 2., 3.]
#     assert exc.value.args[0] == "model_names has not been set"
#
#
# def test_single():
#     s = SED()
#     s.model_names = ['a', 'b', 'c', 'd', 'e']  # this sets the number of models
#     s.apertures = None  # this sets the number of apertures
#     s.flux = [1., 2., 3., 4., 5.]
#     s.error = [0.1, 0.2, 0.3, 0.4, 0.5]
#
#
# def test_single_invalid_lengths():
#
#     s = SED()
#     s.apertures = None
#     s.model_names = ['a', 'b', 'c', 'd']
#
#     with pytest.raises(ValueError) as exc:
#         s.flux = [1., 2., 3.]
#     assert exc.value.args[0] == 'flux has incorrect length (expected 4 but found 3)'
#
#     with pytest.raises(ValueError) as exc:
#         s.error = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
#     assert exc.value.args[0] == 'error has incorrect length (expected 4 but found 6)'
#
#
# @pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
# def test_single_invalid_types(value):
#
#     s = SED()
#
#     s.apertures = None
#     s.model_names = ['a', 'b', 'c', 'd']
#
#     with pytest.raises(TypeError) as exc:
#         s.flux = value
#     assert exc.value.args[0] == 'flux should be a 1-d sequence'
#
#     with pytest.raises(TypeError) as exc:
#         s.error = value
#     assert exc.value.args[0] == 'error should be a 1-d sequence'
#
#
# def test_single_io_roundtrip(tmpdir):
#
#     c1 = SED()
#     c1.model_names = ['a', 'b', 'c', 'd', 'e']
#     c1.apertures = None
#     c1.flux = [1., 2., 3., 4., 5.]
#     c1.error = [0.1, 0.2, 0.3, 0.4, 0.5]
#
#     filename = str(tmpdir.join('test_single.fits'))
#
#     c1.write(filename)
#     c2 = SED.read(filename)
#
#     assert c1 == c2
#
#
# def test_single_interpolate(tmpdir):
#
#     c1 = SED()
#     c1.model_names = ['a', 'b', 'c', 'd', 'e']
#     c1.apertures = None
#     c1.flux = [1., 2., 3., 4., 5.]
#     c1.error = [0.1, 0.2, 0.3, 0.4, 0.5]
#
#     c2 = c1.interpolate([0.5, 1.0, 1.5])
#
#     assert np.all(c1.model_names == c2.model_names)
#
#     assert np.all(c2.apertures == [0.5, 1.0, 1.5])
#
#     assert np.all(c2.flux[:, 0] == c1.flux[:])
#     assert np.all(c2.flux[:, 1] == c1.flux[:])
#     assert np.all(c2.flux[:, 2] == c1.flux[:])
#
#     assert np.all(c2.error[:, 0] == c1.error[:])
#     assert np.all(c2.error[:, 1] == c1.error[:])
#     assert np.all(c2.error[:, 2] == c1.error[:])
#
#
# def test_multiple():
#     s = SED()
#     s.model_names = ['a', 'b', 'c']
#     s.apertures = [1., 2.]
#     s.flux = [[1., 2.], [3., 4.], [5., 6.]]
#     s.error = [[0.1, 0.2], [0.1, 0.2], [0.1, 0.2]]
#
#
# def test_multiple_invalid_lengths():
#
#     s = SED()
#
#     s.apertures = [1., 2.]
#     s.model_names = ['a', 'b', 'c']
#
#     with pytest.raises(ValueError) as exc:
#         s.flux = np.ones((5, 2))
#     assert exc.value.args[0] == 'flux has incorrect shape (expected (3, 2) but found (5, 2))'
#
#     with pytest.raises(ValueError) as exc:
#         s.error = np.ones((3, 3))
#     assert exc.value.args[0] == 'error has incorrect shape (expected (3, 2) but found (3, 3))'
#
#
# @pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
# def test_multiple_invalid_types(value):
#
#     s = SED()
#
#     s.apertures = [1., 2.]
#     s.model_names = ['a', 'b', 'c']
#
#     with pytest.raises(TypeError) as exc:
#         s.flux = value
#     assert exc.value.args[0] == 'flux should be a 2-d array'
#
#     with pytest.raises(TypeError) as exc:
#         s.error = value
#     assert exc.value.args[0] == 'error should be a 2-d array'
#
#
# def test_multiple_roundtrip(tmpdir):
#
#     c1 = SED()
#     c1.model_names = ['a', 'b', 'c']
#     c1.apertures = [1., 2.]
#     c1.flux = [[1., 2.], [3., 4.], [5., 6.]]
#     c1.error = [[0.1, 0.2], [0.1, 0.2], [0.1, 0.2]]
#
#     filename = str(tmpdir.join('test_multiple.fits'))
#
#     c1.write(filename)
#     c2 = SED.read(filename)
#
#     assert c1 == c2
#
#
# def test_multiple_interpolate():
#
#     c1 = SED()
#     c1.model_names = ['a', 'b']
#     c1.apertures = [1., 2., 3.]
#     c1.flux = [[1., 2., 3.], [4., 5., 6.]]
#     c1.error = [[0.1, 0.2, 0.4], [0.5, 0.3, 0.1]]
#
#     c2 = c1.interpolate([1.5, 2.5])
#
#     assert np.all(c1.model_names == c2.model_names)
#
#     assert_array_almost_equal_nulp(c2.apertures, [1.5, 2.5], 10)
#
#     assert_array_almost_equal_nulp(c2.flux[:, 0], np.array([1.5, 4.5]), 10)
#     assert_array_almost_equal_nulp(c2.flux[:, 1], np.array([2.5, 5.5]), 10)
#
#     assert_array_almost_equal_nulp(c2.error[:, 0], np.array([0.15, 0.4]), 10)
#     assert_array_almost_equal_nulp(c2.error[:, 1], np.array([0.3, 0.2]), 10)




def test_read_single_aperture():
    # This actually tests a lot of internal type/dimension checking for the
    # properties.
    s = SED.read(os.path.join(ROOT,'data','kt09250g+4.0z-0.5_sed.fits.gz'))
    assert s.n_wav == 1321
    assert s.n_ap == 1


def test_read_multiple_aperture():
    # This actually tests a lot of internal type/dimension checking for the
    # properties.
    s = SED.read(os.path.join(ROOT,'data','3000030_1_sed.fits.gz'))
    assert s.n_wav == 250
    assert s.n_ap == 50


def test_roundtrip_single_aperture(tmpdir):
    s = SED.read(os.path.join(ROOT,'data','kt09250g+4.0z-0.5_sed.fits.gz'))
    temp_file = tmpdir.join('test_roundtrip_single_aperture').strpath
    s.write(temp_file)
    s2 = SED.read(temp_file)
    assert s == s2


def test_roundtrip_multiple_aperture(tmpdir):
    s = SED.read(os.path.join(ROOT,'data','3000030_1_sed.fits.gz'))
    temp_file = tmpdir.join('test_roundtrip_multipleaperture').strpath
    s.write(temp_file)
    s2 = SED.read(temp_file)
    assert s == s2



from __future__ import print_function, division

import os

import numpy as np
from astropy.table import Table
from astropy import units as u

from .convolved_fluxes import ConvolvedFluxes
from . import fitting_routines as f
from .utils import parfile
from .utils.validator import validate_array


class Models(object):

    def __init__(self):

        self.names = None
        self.fluxes = None
        self.distances = None
        self.apertures = None
        self.logd = None
        self.wavelengths = None
        self.distances = None
        self.extended = []

    @property
    def wavelengths(self):
        """
        The wavelengths at which the models are defined
        """
        return self._wavelengths

    @wavelengths.setter
    def wavelengths(self, value):
        if value is None:
            self._wavelengths = None
        else:
            self._wavelengths = validate_array('wavelengths', value, domain='positive', ndim=1, physical_type='length')

    @property
    def distances(self):
        """
        The distances at which the models are defined
        """
        return self._distances

    @distances.setter
    def distances(self, value):
        if value is None:
            self._distances = None
        else:
            self._distances = validate_array('distances', value, domain='positive', ndim=1, physical_type='length')

    @property
    def apertures(self):
        """
        The apertures at which the fluxes are defined
        """
        return self._apertures

    @apertures.setter
    def apertures(self, value):
        if value is None:
            self._apertures = None
        else:
            self._apertures = validate_array('apertures', value, domain='positive', ndim=1, physical_type='length')

    @property
    def fluxes(self):
        """
        The model fluxes
        """
        return self._fluxes

    @fluxes.setter
    def fluxes(self, value):
        if value is None:
            self._fluxes = value
        else:
            if self.n_distances is None:
                self._fluxes = validate_array('fluxes', value, ndim=2,
                                              shape=(self.n_models, self.n_wav),
                                              physical_type=('power', 'flux', 'spectral flux density'))
            else:
                self._fluxes = validate_array('fluxes', value, ndim=3,
                                              shape=(self.n_models, self.n_distances, self.n_wav),
                                              physical_type=('power', 'flux', 'spectral flux density'))

    @property
    def n_ap(self):
        if self.apertures is None:
            return 1
        else:
            return len(self.apertures)

    @property
    def n_wav(self):
        if self.wavelengths is None:
            return None
        else:
            return len(self.wavelengths)

    @property
    def n_distances(self):
        if self.distances is None:
            return None
        else:
            return len(self.distances)

    @property
    def n_models(self):
        if self.names is None:
            return None
        else:
            return len(self.names)

    @property
    def valid(self):
        if self.fluxes is None:
            return None
        else:
            return self.fluxes != 0

    @property
    def log_fluxes_mJy(self):
        values = np.zeros(self.fluxes.shape)
        values[~self.valid] = -np.inf
        values[self.valid] = np.log10(self.fluxes[self.valid].to(u.mJy).value)
        return values

    @classmethod
    def read(cls, directory, filters, distance_range=None, remove_resolved=False):

        m = cls()

        # Read in model parameters
        modpar = parfile.read("%s/models.conf" % directory, 'conf')

        print(" ------------------------------------------------------------")
        print("  => Model parameters")
        print(" ------------------------------------------------------------")
        print("")
        print("   Models              :  %s" % modpar['name'])
        print("   Log[d] stepping     :  %g" % modpar['logd_step'])

        if modpar['aperture_dependent']:

            distance_range_kpc = distance_range.to(u.kpc).value

            if distance_range:
                if distance_range_kpc[0] == distance_range_kpc[1]:
                    n_distances = 1
                    m.distances = np.array([distance_range_kpc[0]]) * u.kpc
                else:
                    n_distances = 1 + (np.log10(distance_range_kpc[1]) - np.log10(distance_range_kpc[0])) / modpar['logd_step']
                    m.distances = np.logspace(np.log10(distance_range_kpc[0]), np.log10(distance_range_kpc[1]), n_distances) * u.kpc
                print("   Number of distances :  %i" % m.n_distances)
            else:
                raise Exception("For aperture-dependent models, a distange range is required")

        print("")
        print(" ------------------------------------------------------------")
        print("  => Reading in convolved fluxes")
        print(" ------------------------------------------------------------")
        print("")

        m.wavelengths = np.zeros(len(filters)) * u.micron

        for ifilt, filt in enumerate(filters):

            filename = '%s/convolved/%s.fits' % (directory, filt['name'])

            if not os.path.exists(filename):
                if os.path.exists(filename + '.gz'):
                    filename += '.gz'
                else:
                    raise Exception("File not found: " + filename)

            print("   Reading " + filename)

            conv = ConvolvedFluxes.read(filename)

            if ifilt == 0:
                if m.n_distances is None:
                    model_fluxes = np.zeros((conv.n_models, len(filters))) * u.mJy
                    extended = None
                else:
                    model_fluxes = np.zeros((conv.n_models, m.n_distances, len(filters))) * u.mJy
                    extended = np.zeros((conv.n_models, m.n_distances, len(filters)), dtype=bool)

            m.wavelengths[ifilt] = conv.wavelength

            if m.n_distances is not None:
                apertures_au = filt['aperture_arcsec'] * m.distances.to(u.pc).value * u.au
                conv = conv.interpolate(apertures_au)
                conv.flux = conv.flux * (u.kpc / m.distances) ** 2
                m.logd = np.log10(m.distances.to(u.kpc).value)
                if remove_resolved:
                    extended[:, :, ifilt] = apertures_au[np.newaxis,:] < conv.radius_sigma_50[:, np.newaxis]
                model_fluxes[:, :, ifilt] = conv.flux
            else:
                model_fluxes[:, ifilt] = conv.flux[:, 0]

        try:
            m.names = np.char.strip(conv.model_names)
        except:
            m.names = np.array([x.strip() for x in conv.model_names], dtype=conv.model_names.dtype)

        m.fluxes = model_fluxes

        if extended is not None:
            m.extended = extended

        return m

    def fit(self, source, av_law, sc_law, av_min, av_max):

        weight, log_flux, log_error = source.get_log_fluxes()

        model_fluxes = self.log_fluxes_mJy

        if model_fluxes.ndim == 2:  # Aperture-independent fitting

            # Use 2-parameter linear regression to find the best-fit av and scale for each model
            residual = log_flux - model_fluxes
            av_best, sc_best = f.linear_regression(residual, weight, av_law, sc_law)

            # Use optimal scaling for Avs that are outside range
            reset1 = (av_best < av_min)
            reset2 = (av_best > av_max)
            av_best[reset1] = av_min
            av_best[reset2] = av_max
            reset = reset1 | reset2
            sc_best[reset] = f.optimal_scaling(residual[reset] - av_best[reset][:, np.newaxis] * av_law[np.newaxis, :], weight, sc_law)

            # Compute best-fit model in each case
            model = av_best[:, np.newaxis] * av_law[np.newaxis, :] + sc_best[:, np.newaxis] * sc_law[np.newaxis,:]

            # Calculate the chi-squared value
            ch_best = f.chi_squared(source.valid, residual, log_error, weight, model)

            # Extract convolved model fluxes for best-fit
            model_fluxes = model + model_fluxes

        elif model_fluxes.ndim == 3:  # Aperture dependent fitting

            # Use optimal scaling to fit the Av
            residual = log_flux - model_fluxes
            av_best = f.optimal_scaling(residual, weight, av_law)

            # Reset to valid range
            av_best[av_best < av_min] = av_min
            av_best[av_best > av_max] = av_max

            # Compute best-fit model in each case
            model = av_best[:, :, np.newaxis] * av_law[np.newaxis, np.newaxis,:]

            # Calculate the chi-squared value
            ch_best = f.chi_squared(source.valid, residual, log_error, weight, model)

            # Remove extended objects
            if type(self.extended) == np.ndarray:
                reset = np.any(self.extended[:, :, source.valid > 0], axis=2)
                ch_best[reset] = np.inf

            # Find best-fit distance in each case
            best = np.argmin(ch_best, axis=1)

            sc_best = self.logd[best]

            ch_best = ch_best[np.arange(self.n_models), best]
            av_best = av_best[np.arange(self.n_models), best]

            # Extract convolved model fluxes for best-fit
            model_fluxes = (model + model_fluxes)[np.arange(self.n_models), best, :]

        else:

            raise Exception("Unexpected number of dimensions in flux array")

        return av_best, sc_best, ch_best, self.names, model_fluxes


def load_parameter_table(model_dir):

    if os.path.exists(model_dir + '/parameters.fits'):
        t = Table.read(model_dir + '/parameters.fits')
    elif os.path.exists(model_dir + '/parameters.fits.gz'):
        t = Table.read(model_dir + '/parameters.fits.gz')
    else:
        raise Exception("Parameter file not found in %s" % model_dir)

    return t



import os
import glob

from astropy import units as u

from sedfitter import (fit, plot, plot_params_1d, plot_params_2d,
                       write_parameters, write_parameter_ranges)
from sedfitter.extinction import Extinction

# Define path to models
model_dir = '/Volumes/Data/models/models_r06'

# Read in extinction law)
extinction = Extinction.from_file('kmh94.par', columns=[0, 3],
                                  wav_unit=u.micron, chi_unit=u.cm**2 / u.g)

# Define filters and apertures
filters = ['2J', '2H', '2K', 'I1', 'I2', 'I3', 'I4']
apertures = [3., 3., 3., 3., 3., 3., 3.] * u.arcsec

# Run the fitting
fit('data_glimpse', filters, apertures, model_dir,
    'output.fitinfo',
    extinction_law=extinction,
    distance_range=[1., 2.] * u.kpc,
    av_range=[0., 40.])

# For the remaining commands, we always select the models with chi^2-chi_best^2
# per datapoint less than 3.
select_format = ('F', 3)

# Make SED plots
plot('output.fitinfo', 'plots_seds', plot_max=100, select_format=select_format)

# Make histograms of the disk mass
plot_params_1d('output.fitinfo', 'MDISK', 'plots_mdisk',
               log_x=True, select_format=select_format)

# Make 2-d plots of the envelope infall rate vs disk mass
plot_params_2d('output.fitinfo', 'MDISK', 'MDOT', 'plots_mdot_mdisk',
               log_x=True, log_y=True, select_format=select_format)

# Write out all models with a delta chi^2-chi_best^2 per datapoint < 3
write_parameters('output.fitinfo', 'parameters.txt',
                 select_format=select_format)

# Write out the min/max ranges corresponding to the above file
write_parameter_ranges('output.fitinfo', 'parameter_ranges.txt',
                       select_format=select_format)


import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal_nulp

from astropy import units as u

from .. import ConvolvedFluxes

def assert_allclose_quantity(q1, q2):
    np.testing.assert_allclose(q1.value, q2.to(q1.unit).value)


def test_init():
    ConvolvedFluxes()


def test_wavelength():
    c = ConvolvedFluxes()
    c.wavelength = 3.14 * u.micron


@pytest.mark.parametrize('value', ['string', object(), np.array([1, 2, 3])])
def test_wavelength_invalid_type(value):
    c = ConvolvedFluxes()
    with pytest.raises(TypeError) as exc:
        c.wavelength = value
    assert exc.value.args[0] == 'wavelength should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.Jy])
def test_wavelength_invalid_unit(unit):
    c = ConvolvedFluxes()
    with pytest.raises(TypeError) as exc:
        c.wavelength = [1., 2., 3.] * unit
    assert exc.value.args[0] == 'wavelength should be given in units of length'


def test_aperture_none():
    c = ConvolvedFluxes()
    c.apertures = None


def test_aperture_list():
    c = ConvolvedFluxes()
    c.apertures = [1., 2., 3.] * u.au


@pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
def test_aperture_invalid_type(value):

    c = ConvolvedFluxes()

    with pytest.raises(TypeError) as exc:
        c.apertures = value
    assert exc.value.args[0] == 'apertures should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.Jy])
def test_aperture_invalid_unit(unit):

    c = ConvolvedFluxes()

    with pytest.raises(TypeError) as exc:
        c.apertures = [1., 2., 3.] * unit
    assert exc.value.args[0] == 'apertures should be given in units of length'


def test_model_name_list():
    c = ConvolvedFluxes()
    c.model_names = ['a', 'b', 'c']


@pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
def test_model_name_invalid_type(value):

    c = ConvolvedFluxes()

    with pytest.raises(TypeError) as exc:
        c.model_names = value
    assert exc.value.args[0] == 'model_names should be a 1-d sequence'


def test_model_name_unset():

    c = ConvolvedFluxes()

    with pytest.raises(ValueError) as exc:
        c.flux = [1., 2., 3.] * u.mJy
    assert exc.value.args[0] == "model_names has not been set"

    with pytest.raises(ValueError) as exc:
        c.error = [1., 2., 3.] * u.mJy
    assert exc.value.args[0] == "model_names has not been set"


def test_single():
    c = ConvolvedFluxes()
    c.model_names = ['a', 'b', 'c', 'd', 'e']  # this sets the number of models
    c.apertures = None  # this sets the number of apertures
    c.flux = [(1.,), (2.,), (3.,), (4.,), (5.,)] * u.mJy
    c.error = [(0.1,), (0.2,), (0.3,), (0.4,), (0.5,)] * u.mJy


def test_single_invalid_lengths():

    c = ConvolvedFluxes()
    c.apertures = None
    c.model_names = ['a', 'b', 'c', 'd']

    with pytest.raises(ValueError) as exc:
        c.flux = [(1.,), (2.,), (3.,)] * u.mJy
    assert exc.value.args[0] == 'flux has incorrect shape (expected (4, 1) but found (3, 1))'

    with pytest.raises(ValueError) as exc:
        c.error = [(0.1,), (0.2,), (0.3,), (0.4,), (0.5,), (0.6,)] * u.mJy
    assert exc.value.args[0] == 'error has incorrect shape (expected (4, 1) but found (6, 1))'


@pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
def test_single_invalid_types(value):

    c = ConvolvedFluxes()

    c.apertures = None
    c.model_names = ['a', 'b', 'c', 'd']

    with pytest.raises(TypeError) as exc:
        c.flux = value
    assert exc.value.args[0] == 'flux should be given as a Quantity object'

    with pytest.raises(TypeError) as exc:
        c.error = value
    assert exc.value.args[0] == 'error should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.m])
def test_single_invalid_units(unit):

    c = ConvolvedFluxes()

    c.apertures = None
    c.model_names = ['a', 'b', 'c', 'd']

    with pytest.raises(TypeError) as exc:
        c.flux = [(1.,), (2.,), (3.,), (4.,)] * unit
    assert exc.value.args[0] == 'flux should be given in units of power, flux, spectral flux density'

    with pytest.raises(TypeError) as exc:
        c.error = [(1.,), (2.,), (3.,), (4.,)] * unit
    assert exc.value.args[0] == 'error should be given in units of power, flux, spectral flux density'

def test_single_io_roundtrip(tmpdir):

    c1 = ConvolvedFluxes()
    c1.model_names = ['a', 'b', 'c', 'd', 'e']
    c1.apertures = None
    c1.flux = [(1.,), (2.,), (3.,), (4.,), (5.,)] * u.mJy
    c1.error = [(0.1,), (0.2,), (0.3,), (0.4,), (0.5,)] * u.mJy

    filename = str(tmpdir.join('test_single.fits'))

    c1.write(filename)
    c2 = ConvolvedFluxes.read(filename)

    assert c1 == c2


def test_single_interpolate(tmpdir):

    c1 = ConvolvedFluxes()
    c1.model_names = ['a', 'b', 'c', 'd', 'e']
    c1.apertures = None
    c1.flux = [(1.,), (2.,), (3.,), (4.,), (5.,)] * u.mJy
    c1.error = [(0.1,), (0.2,), (0.3,), (0.4,), (0.5,)] * u.mJy

    c2 = c1.interpolate([0.5, 1.0, 1.5] * u.au)

    assert np.all(c1.model_names == c2.model_names)

    assert np.all(c2.apertures == [0.5, 1.0, 1.5] * u.au)

    assert np.all(c2.flux[:, 0] == c1.flux[:, 0])
    assert np.all(c2.flux[:, 1] == c1.flux[:, 0])
    assert np.all(c2.flux[:, 2] == c1.flux[:, 0])

    assert np.all(c2.error[:, 0] == c1.error[:, 0])
    assert np.all(c2.error[:, 1] == c1.error[:, 0])
    assert np.all(c2.error[:, 2] == c1.error[:, 0])


def test_multiple():
    c = ConvolvedFluxes()
    c.model_names = ['a', 'b', 'c']
    c.apertures = [1., 2.] * u.au
    c.flux = [[1., 2.], [3., 4.], [5., 6.]] * u.mJy
    c.error = [[0.1, 0.2], [0.1, 0.2], [0.1, 0.2]] * u.mJy


def test_multiple_invalid_lengths():

    c = ConvolvedFluxes()

    c.apertures = [1., 2.] * u.au
    c.model_names = ['a', 'b', 'c']

    with pytest.raises(ValueError) as exc:
        c.flux = np.ones((5, 2)) * u.mJy
    assert exc.value.args[0] == 'flux has incorrect shape (expected (3, 2) but found (5, 2))'

    with pytest.raises(ValueError) as exc:
        c.error = np.ones((3, 3)) * u.mJy
    assert exc.value.args[0] == 'error has incorrect shape (expected (3, 2) but found (3, 3))'


@pytest.mark.parametrize('value', ['string', 1, 0.5, np.zeros((2, 2, 3))])
def test_multiple_invalid_types(value):

    c = ConvolvedFluxes()

    c.apertures = [1., 2.] * u.au
    c.model_names = ['a', 'b', 'c']

    with pytest.raises(TypeError) as exc:
        c.flux = value
    assert exc.value.args[0] == 'flux should be given as a Quantity object'

    with pytest.raises(TypeError) as exc:
        c.error = value
    assert exc.value.args[0] == 'error should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.m])
def test_multiple_invalid_unit(unit):

    c = ConvolvedFluxes()

    c.apertures = [1., 2.] * u.au
    c.model_names = ['a', 'b', 'c']

    with pytest.raises(TypeError) as exc:
        c.flux = [[1., 2.], [3., 4.], [5., 6.]] * unit
    assert exc.value.args[0] == 'flux should be given in units of power, flux, spectral flux density'

    with pytest.raises(TypeError) as exc:
        c.error = [[1., 2.], [3., 4.], [5., 6.]] * unit
    assert exc.value.args[0] == 'error should be given in units of power, flux, spectral flux density'


def test_multiple_roundtrip(tmpdir):

    c1 = ConvolvedFluxes()
    c1.model_names = ['a', 'b', 'c']
    c1.apertures = [1., 2.] * u.au
    c1.flux = [[1., 2.], [3., 4.], [5., 6.]] * u.mJy
    c1.error = [[0.1, 0.2], [0.1, 0.2], [0.1, 0.2]] * u.mJy

    filename = str(tmpdir.join('test_multiple.fits'))

    c1.write(filename)
    c2 = ConvolvedFluxes.read(filename)

    assert c1 == c2


def test_multiple_interpolate():

    c1 = ConvolvedFluxes()
    c1.model_names = ['a', 'b']
    c1.apertures = [1., 2., 3.] * u.au
    c1.flux = [[1., 2., 3.], [4., 5., 6.]] * u.mJy
    c1.error = [[0.1, 0.2, 0.4], [0.5, 0.3, 0.1]] * u.mJy

    c2 = c1.interpolate([1.5, 2.5] * u.au)

    assert np.all(c1.model_names == c2.model_names)

    assert_allclose_quantity(c2.apertures, [1.5, 2.5] * u.au)

    assert_allclose_quantity(c2.flux[:, 0], np.array([1.5, 4.5]) * u.mJy)
    assert_allclose_quantity(c2.flux[:, 1], np.array([2.5, 5.5]) * u.mJy)

    assert_allclose_quantity(c2.error[:, 0], np.array([0.15, 0.4]) * u.mJy)
    assert_allclose_quantity(c2.error[:, 1], np.array([0.3, 0.2]) * u.mJy)



from __future__ import print_function, division

import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
from copy import deepcopy

from astropy.table import Table
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
from matplotlib.patches import Polygon
from matplotlib.ticker import LogFormatterMathtext

from .fit_info import FitInfo
from .extinction import Extinction
from .models import load_parameter_table
from .utils import io
from .utils.formatter import LogFormatterMathtextAuto


plt.rc('text', usetex=False)
plt.rc('axes', titlesize='small')
plt.rc('axes', labelsize='small')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('font', family='serif')
plt.rc('axes', linewidth=0.5)
plt.rc('patch', linewidth=0.5)

fp = FontProperties(size='small')


def get_axes(fig):
    vxmin, vxmax = 1.5, 5.5
    vymin, vymax = 1.0, 4.0
    width, height = fig.get_figwidth(), fig.get_figheight()
    rect = [vxmin / width, vymin / width, (vxmax - vxmin) / width, (vymax - vymin) / height]
    return fig.add_axes(rect)


def plot_params_1d(input_file, parameter, output_dir=None,
                   select_format=("N", 1), log_x=False, log_y=True,
                   label=None, bins=30, additional={}, plot_name=True,
                   format='pdf'):
    """
    Make histogram plots of parameters

    Parameters
    ----------
    input_file : str
        File containing the fit information
    parameter : str
        The parameter to plot a histogram of
    output_dir : str, optional
        If specified, plots are written to that directory
    select_format : tuple, optional
        Tuple specifying which fits should be plotted. See the documentation
        for a description of the tuple syntax.
    log_x : bool, optional
        Whether to plot the x-axis values in log space
    log_y : bool, optional
        Whether to plot the y-axis values in log space
    label : str, optional
        The x-axis label (if not specified, the parameter name is used)
    bins : int, optional
        The number of bins for the histogram
    additional : dict, optional
        A dictionary specifying additional parameters not listed in the
        parameter list for the models. Each item of the dictionary should
        itself be a dictionary giving the values for each model (where the key
        is the model name).
    plot_name: bool, optional
        Whether to show the source name on the plot(s).
    format: str, optional
        The file format to use for the plot, if output_dir is specified.

    """

    if output_dir is None:
        raise ValueError("No output directory has been specified")
    # Create output directory
    io.create_dir(output_dir)

    # Open output file
    fin = open(input_file, 'rb')

    # Read in header of output file
    model_dir = pickle.load(fin)
    filters = pickle.load(fin)
    extinction = pickle.load(fin)

    # Read in table of parameters for model grid
    t = load_parameter_table(model_dir)

    # Sort alphabetically
    t['MODEL_NAME'] = np.char.strip(t['MODEL_NAME'])
    t.sort('MODEL_NAME')
    tpos = deepcopy(t)
    if log_x:
        tpos = tpos[tpos[parameter] > 0.]

    # Initialize figure
    fig = plt.figure()
    ax = get_axes(fig)

    # Find range of values
    pmin, pmax = tpos[parameter].min(), tpos[parameter].max()

    # Compute histogram
    if log_x:
        hist_all, edges = np.histogram(np.log10(tpos[parameter]), bins=bins, range=[np.log10(pmin), np.log10(pmax)])
        center = (edges[1:] + edges[:-1]) / 2.
        edges, center = 10. ** edges, 10. ** center
    else:
        hist_all, edges = np.histogram(t[parameter], bins=bins, range=[pmin, pmax])
        center = (edges[1:] + edges[:-1]) / 2.

    # Grayscale showing all models
    p = []
    for i in range(len(hist_all)):
        p.append((edges[i], max(hist_all[i], 0.01)))
        p.append((edges[i + 1], max(hist_all[i], 0.01)))
    p.append((edges[-1], 0.01))
    p.append((edges[0], 0.01))

    p = Polygon(p, facecolor='0.8', edgecolor='none')
    ax.add_patch(p)

    ax.set_xlabel(parameter if label is None else label)

    if log_x:
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(LogFormatterMathtextAuto())
    if log_y:
        ax.set_yscale('log')
        ax.yaxis.set_major_formatter(LogFormatterMathtextAuto())

    ax.set_xlim(pmin, pmax)
    ax.set_ylim(0.1, hist_all.max() * 10.)

    ax.set_autoscale_on(False)

    pfits = None
    source_label = None

    while True:  # Loop over the fits

        # Read in next fit
        try:
            info = pickle.load(fin)
        except EOFError:
            break

        # Remove previous histogram
        if pfits is not None:
            pfits.remove()
        if source_label is not None:
            source_label.remove()

        # Filter fits
        info.keep(select_format[0], select_format[1])

        # Get filtered and sorted table of parameters
        tsorted = info.filter_table(t, additional=additional)

        # Compute histogram
        if log_x:
            hist, _ = np.histogram(np.log10(tsorted[parameter]), bins=bins, range=[np.log10(pmin), np.log10(pmax)])
        else:
            hist, _ = np.histogram(tsorted[parameter], bins=bins, range=[pmin, pmax])

        # Histogram showing values for good-fitting models
        p = []
        for i in range(len(hist)):
            p.append((edges[i], max(hist[i], 0.01)))
            p.append((edges[i + 1], max(hist[i], 0.01)))
        p.append((edges[-1], 0.01))
        p.append((edges[0], 0.01))
        pfits = Polygon(p, hatch='/', edgecolor='black', facecolor='none')
        ax.add_patch(pfits)

        if plot_name:
            source_label = ax.text(0.5, 0.95, info.source.name,
                                   horizontalalignment='center',
                                   verticalalignment='center',
                                   transform=ax.transAxes,
                                   fontproperties=fp, zorder=200)

        # Save to file
        filename = "%s/%s.%s" % (output_dir, info.source.name, format)
        fig.savefig(filename, bbox_inches='tight')
        plt.close(fig)

    # Close input and output files
    fin.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import print_function, division

from .fit import fit
from .plot import plot, plot_source_data
from .plot_params_1d import plot_params_1d
from .plot_params_2d import plot_params_2d
from .filter_output import filter_output
from .extract_parameters import extract_parameters
from .write_parameters import write_parameters
from .write_parameter_ranges import write_parameter_ranges

from . import convolve
from . import filter
from . import utils

__version__ = '0.1.2.dev'



from __future__ import print_function, division

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np
from astropy import units as u

from ..utils.validator import validate_array


class Extinction(object):

    def __init__(self):
        self.wav = None
        self.chi = None

    @property
    def wav(self):
        return self._wav

    @wav.setter
    def wav(self, value):
        if value is None:
            self._wav = None
        else:
            self._wav = validate_array('wav', value, ndim=1,
                                       shape=None if self.chi is None else self.chi.shape,
                                       physical_type='length')

    @property
    def chi(self):
        return self._chi

    @chi.setter
    def chi(self, value):
        if value is None:
            self._chi = None
        else:
            self._chi = validate_array('chi', value, ndim=1,
                                       shape=None if self.wav is None else self.wav.shape,
                                       physical_type='area per unit mass')

    @classmethod
    def from_file(cls, filename, columns=(0, 1),
                  wav_unit=u.micron, chi_unit=u.cm ** 2 / u.g):
        """
        Read an extinction law from an ASCII file.

        This reads in two columns: the wavelength, and the opacity (in units
        of area per unit mass).

        Parameters
        ----------
        filename : str, optional
            The name of the file to read the extinction law from
        columns : tuple or list, optional
            The columns to use for the wavelength and opacity respectively
        wav_unit : :class:`~astropy.units.core.Unit`
            The units to assume for the wavelength
        chi_unit : :class:`~astropy.units.core.Unit`
            The units to assume for the opacity
        """

        self = cls()

        f = np.loadtxt(filename, dtype=[('wav', float), ('chi', float)],
                       usecols=columns)

        self.wav = f['wav'] * wav_unit
        self.chi = f['chi'] * chi_unit

        return self

    def get_av(self, wav):
        """
        Interpolate the Av at given wavelengths

        Parameters
        ----------
        wav : :class:`~astropy.units.quantity.Quantity`
            The wavelengths at which to interpolate the visual extinction.
        """
        if isinstance(wav, u.Quantity) and wav.unit.is_equivalent(u.m):
            return (-0.4 * np.interp(wav.to(self.wav.unit), self.wav, self.chi, left=0., right=0.)
                    / np.interp(([0.55] * u.micron).to(self.wav.unit), self.wav, self.chi))
        else:
            raise TypeError("wav should be given as a Quantity object with units of length")

    @classmethod
    def from_table(cls, table):
        self = cls()
        self.wav = table['wav'].data * table['wav'].unit
        self.chi = table['chi'].data * table['chi'].unit
        return self

    def to_table(self):
        from astropy.table import Table, Column
        from astropy import units as u
        t = Table()
        t['wav'] = self.wav
        t['chi'] = self.chi
        return t

    def __getstate__(self):
        return {
            'wav': self.wav,
            'chi': self.chi,
        }

    def __setstate__(self, d):
        self.__init__()
        self.wav = d['wav']
        self.chi = d['chi']



from __future__ import print_function, division

import time

import numpy as np


class Timer(object):

    def __init__(self):
        self.time1 = time.time()
        self.n = 0
        self.step = 1
        print("   # Sources    CPU time (sec)    Sources/sec  ")
        print(" ----------------------------------------------")

    def display(self, force=False):
        self.n += 1
        if np.mod(self.n, self.step) == 0:
            self.time2 = time.time()
            if self.time2 - self.time1 < 1.:
                self.step *= 10
            else:
                print("    %7i       %10.1f        %7.2f" % (self.n, self.time2 - self.time1, self.n / (self.time2 - self.time1)))
        elif force:
            print("    %7i       %10.1f        %7.2f" % (self.n, self.time2 - self.time1, self.n / (self.time2 - self.time1)))



#!/usr/bin/env python

from setuptools import setup, Command

from distutils.command.build_py import build_py

class SEDFitterTest(Command):

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        import os
        import shutil
        import tempfile

        # First ensure that we build the package so that 2to3 gets executed
        self.reinitialize_command('build', inplace=False)
        self.run_command('build')
        build_cmd = self.get_finalized_command('build')
        new_path = os.path.abspath(build_cmd.build_lib)

        # Copy the build to a temporary directory for the purposes of testing
        # - this avoids creating pyc and __pycache__ directories inside the
        # build directory
        tmp_dir = tempfile.mkdtemp(prefix='sedfitter-test-')
        testing_path = os.path.join(tmp_dir, os.path.basename(new_path))
        shutil.copytree(new_path, testing_path)

        import sys
        import subprocess

        errno = subprocess.call([sys.executable, os.path.abspath('runtests.py')], cwd=testing_path)
        raise SystemExit(errno)

setup(name='sedfitter',
      version='0.9.0',
      description='SED Fitter in Python',
      author='Thomas Robitaille',
      author_email='thomas.robitaille@gmail.com',
      packages=['sedfitter',
                'sedfitter.convolve',
                'sedfitter.convolved_fluxes',
                'sedfitter.convolved_fluxes.tests',
                'sedfitter.extinction',
                'sedfitter.extinction.tests',
                'sedfitter.filter',
                'sedfitter.sed',
                'sedfitter.sed.tests',
                'sedfitter.source',
                'sedfitter.source.tests',
                'sedfitter.tests',
                'sedfitter.utils',
                'sedfitter.utils.tests'],
      package_data={'sedfitter.sed.tests':['data/*.fits.gz'],
                    'sedfitter.utils.tests':['data/*.conf', 'data/*.par']},
      provides=['sedfitter'],
      requires=['numpy', 'scipy', 'matplotlib', 'astropy'],
      cmdclass={'build_py': build_py, 'test':SEDFitterTest},
      keywords=['Scientific/Engineering'],
     )




from __future__ import print_function, division

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np

from .source import Source


class FitInfo(object):

    def __init__(self, source=None):

        self.source = source
        self.av = None
        self.sc = None
        self.chi2 = None
        self.model_id = None
        self.model_name = None
        self.model_fluxes = None

    def sort(self):

        order = np.argsort(self.chi2)
        self.av = self.av[order]
        self.sc = self.sc[order]
        self.chi2 = self.chi2[order]
        self.model_name = self.model_name[order]
        if self.model_fluxes is not None:
            self.model_fluxes = self.model_fluxes[order, :]
        self.model_id = order

    def keep(self, form, number):

        if form == 'A':
            n_fits = len(self.chi2)
        elif form == 'N' or not form:  # Form N is parsed as boolean
            n_fits = int(number)
        elif form == 'C':
            n_fits = np.sum(self.chi2 <= number)
        elif form == 'D':
            n_fits = np.sum(self.chi2 - self.chi2[0] <= number)
        elif form == 'E':
            n_fits = np.sum((self.chi2 / self.source.n_wav) <= number)
        elif form == 'F':
            n_fits = np.sum((self.chi2 - self.chi2[0]) / self.source.n_wav <= number)
        else:
            raise Exception("Unknown format: %s" % form)

        self.av = self.av[:n_fits]
        self.sc = self.sc[:n_fits]
        self.chi2 = self.chi2[:n_fits]
        self.model_name = self.model_name[:n_fits]
        if self.model_fluxes is not None:
            self.model_fluxes = self.model_fluxes[:n_fits]
        self.model_id = self.model_id[:n_fits]

    def __getattr__(self, attribute):
        if attribute == 'n_fits':
            return len(self.chi2)
        else:
            raise AttributeError(attribute)

    def __getstate__(self):
        return {
            'source': self.source,
            'av': self.av,
            'sc': self.sc,
            'chi2': self.chi2,
            'model_id': self.model_id,
            'model_name': self.model_name,
            'model_fluxes': self.model_fluxes,
        }

    def __setstate__(self, d):
        self.__init__()
        self.source = d['source']
        self.av = d['av']
        self.sc = d['sc']
        self.chi2 = d['chi2']
        self.model_id = d['model_id']
        self.model_name = d['model_name']
        self.model_fluxes = d['model_fluxes']

    def filter_table(self, input_table, additional={}):
        """
        Given an input table, return only the rows matching the FitInfo object, and in the same order.
        """

        if not "MODEL_NAME" in input_table.dtype.names:
            raise ValueError("Input table should contain a MODEL_NAME column")

        subset = np.in1d(input_table['MODEL_NAME'], self.model_name)
        table_subset = input_table[subset]
        index = np.argsort(np.argsort(self.model_name))
        table_sorted = table_subset[index]

        # Double check that the sorting worked
        if not np.all(self.model_name == table_sorted['MODEL_NAME']):
            raise Exception("Parameter file sorting failed")

        # Add additional parameter columns if necessary
        for par in additional:
            if par in table_sorted.columns:
                raise Exception("Parameter {0} already exists in table".format(par))
            table_sorted[par] = np.zeros(len(table_sorted), dtype=float)
            for i, name in enumerate(table_sorted['MODEL_NAME']):
                table_sorted[par][i] = additional[par][name.strip()]

        return table_sorted



import pytest
import numpy as np
from numpy.testing import assert_allclose

from astropy import units as u

from .. import Extinction


def assert_allclose_quantity(q1, q2):
    np.testing.assert_allclose(q1.value, q2.to(q1.unit).value)


def test_extinction_init():
    e = Extinction()


@pytest.mark.parametrize('value', [[1, 0, 1, 1, 2], (0, 1, 2, 3, 4), np.array([1., 2., 3.])])
def test_extinction_wav(value):
    e = Extinction()
    e.wav = value * u.micron


@pytest.mark.parametrize('value', [1, 1., 'a', object()])
def test_extinction_wav_invalid_type(value):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.wav = value
    assert exc.value.args[0] == 'wav should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.Jy])
def test_extinction_wav_invalid_unit(unit):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.wav = [1., 2., 3.] * unit
    assert exc.value.args[0] == 'wav should be given in units of length'


@pytest.mark.parametrize('value', [np.array([[1, 2]]), np.array(3)])
def test_extinction_wav_invalid_shape(value):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.wav = value * u.micron
    assert exc.value.args[0] == 'wav should be a 1-d sequence'


@pytest.mark.parametrize('value', [[1, 0, 1, 1, 2], (0, 1, 2, 3, 4), np.array([1., 2., 3.])])
def test_extinction_chi(value):
    e = Extinction()
    e.chi = value * u.cm ** 2 / u.g


@pytest.mark.parametrize('value', [1, 1., 'a', object()])
def test_extinction_chi_invalid_type(value):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.chi = value
    assert exc.value.args[0] == 'chi should be given as a Quantity object'


@pytest.mark.parametrize('unit', [u.arcsec, u.kg, u.Jy])
def test_extinction_chi_invalid_unit(unit):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.chi = [1., 2., 3.] * unit
    assert exc.value.args[0] == 'chi should be given in units of area per unit mass'


@pytest.mark.parametrize('value', [np.array([[1, 2]]), np.array(3)])
def test_extinction_chi_invalid_shape(value):
    e = Extinction()
    with pytest.raises(TypeError) as exc:
        e.chi = value * u.cm ** 2 / u.g
    assert exc.value.args[0] == 'chi should be a 1-d sequence'


@pytest.mark.parametrize(('attribute'), ['wav', 'chi'])
def test_dimensions_match(attribute):
    units = {}
    units['wav'] = u.micron
    units['chi'] = u.cm ** 2 / u.g
    e = Extinction()
    setattr(e, attribute, [1, 2, 3] * units[attribute])
    for other in ['wav', 'chi']:
        if other != attribute:
            with pytest.raises(ValueError) as exc:
                setattr(e, other, [1, 2, 3, 4] * units[other])
            assert exc.value.args[0] == "{0} has incorrect length (expected 3 but found 4)".format(other)


def test_extinction_table():

    e = Extinction()
    e.wav = [0., 1., 2., 3.3, 5.5] * u.micron
    e.chi = [1., 3., 2., 3., 5.] * u.cm ** 2 / u.g

    table = e.to_table()

    e2 = Extinction.from_table(table)

    assert_allclose_quantity(e.wav, e2.wav)
    assert_allclose_quantity(e.chi, e2.chi)


def test_extinction_numpy_io(tmpdir):

    wav = [0., 1., 2., 3.3, 5.5]
    chi = [1., 3., 2., 3., 5.]

    filename = str(tmpdir.join('test_extinction.fits'))
    np.savetxt(filename, list(zip(wav, chi)))

    e2 = Extinction.from_file(filename)

    assert_allclose(wav, e2.wav)
    assert_allclose(chi, e2.chi)



from __future__ import print_function, division

import os
try:
    import cPickle as pickle
except ImportError:
    import pickle

from astropy.table import Table
import numpy as np

from .fit_info import FitInfo
from .extinction import Extinction
from .models import load_parameter_table


def extract_parameters(input=None, output_prefix=None, output_suffix=None,
                       parameters='all', select_format=("N", 1), header=True):

    fin = open(input, 'rb')

    # Read in header of output file
    model_dir = pickle.load(fin)
    filters = pickle.load(fin)
    extinction_law = pickle.load(fin)

    # Read in table of parameters for model grid
    t = load_parameter_table(model_dir)

    format = {}
    for par in t.dtype.names:
        if par == 'MODEL_NAME':
            format[par] = "%11s"
        else:
            format[par] = "%11.3e"

    if parameters == 'all':
        parameters = t.dtype.names

    while True:

        # Read in next fit
        try:
            info = pickle.load(fin)
        except EOFError:
            break

        # Filter fits
        info.keep(select_format[0], select_format[1])

        output = ""
        if output_prefix:
            output += output_prefix
        output += info.source.name
        if output_suffix:
            output += output_suffix

        fout = open(output, 'w')

        if header:
            fout.write("%11s %11s %11s " % ("CHI2", "AV", "SC"))
            fout.write(" ".join([("%11s" % par) for par in parameters]))
            fout.write("\n")

        # Get filtered and sorted table of parameters
        tsorted = info.filter_table(t)

        for i in range(info.n_fits):

            row = tsorted[i]

            basic = "%11.3e %11.3e %11.3e " % (info.chi2[i], info.av[i], info.sc[i])
            pars = " ".join([(format[par] % row[par]) for par in parameters])

            fout.write(basic + pars + "\n")

        fout.close()

    # Close input and output files
    fin.close()



from __future__ import print_function, division

import os
import glob

import numpy as np
from astropy.io import fits
from astropy.logger import log
from astropy.table import Table
from astropy import units as u
from astropy.utils.console import ProgressBar

from ..convolved_fluxes import ConvolvedFluxes
from ..sed import SED
from ..models import load_parameter_table
from .. import six


def convolve_model_dir_monochromatic(model_dir, overwrite=False, max_ram=8,
                                     wav_min=-np.inf * u.micron, wav_max=np.inf * u.micron):
    """
    Convolve all the model SEDs in a model directory

    Parameters
    ----------
    model_dir: str
        The path to the model directory
    overwrite: bool, optional
        Whether to overwrite the output files
    max_ram: float, optional
        The maximum amount of RAM that can be used (in Gb)
    wav_min: float, optional
        The minimum wavelength to consider. Only wavelengths above this value
        will be output.
    wav_max: float, optional
        The maximum wavelength to consider. Only wavelengths below this value
        will be output.
    """

    # Create 'convolved' sub-directory if needed
    if not os.path.exists(model_dir + '/convolved'):
        os.mkdir(model_dir + '/convolved')

    # Find all SED files to convolve
    sed_files = (glob.glob(model_dir + '/seds/*.fits.gz') +
                 glob.glob(model_dir + '/seds/*/*.fits.gz') +
                 glob.glob(model_dir + '/seds/*.fits') +
                 glob.glob(model_dir + '/seds/*/*.fits'))

    par_table = load_parameter_table(model_dir)

    # Find number of models
    n_models = len(sed_files)

    if n_models == 0:
        raise Exception("No SEDs found in %s" % model_dir)
    else:
        log.info("{0} SEDs found in {1}".format(n_models, model_dir))

    # Find out apertures and wavelengths
    first_sed = SED.read(sed_files[0])
    n_ap = first_sed.n_ap
    apertures = first_sed.apertures
    n_wav = first_sed.n_wav
    wavelengths = first_sed.wav

    # For model grids that are very large, it is not possible to compute all
    # fluxes in one go, so we need to process in chunks in wavelength space.
    chunk_size = min(n_wav, int(np.floor(max_ram * 1024. ** 3 / (4. * 2. * n_models * n_ap))))

    if chunk_size == n_wav:
        log.info("Producing all monochromatic files in one go")
    else:
        log.info("Producing monochromatic files in chunks of {0}".format(chunk_size))

    filters = Table()
    filters['wav'] = wavelengths
    filters['filter'] = np.zeros(wavelengths.shape, dtype='S10')

    # Figure out range of wavelength indices to use
    # (wavelengths array is sorted in reverse order)
    jlo = n_wav - wavelengths[::-1].searchsorted(wav_max)
    jhi = n_wav - wavelengths[::-1].searchsorted(wav_min)
    chunk_size = min(chunk_size, jhi - jlo + 1)

    # Loop over wavelength chunks
    for jmin in range(jlo, jhi, chunk_size):

        # Find upper wavelength to compute
        jmax = min(jmin + chunk_size - 1, jhi)

        log.info('Processing wavelengths {0} to {1}'.format(jmin, jmax))

        # Set up convolved fluxes
        fluxes = [ConvolvedFluxes(model_names=np.zeros(n_models, dtype='U30' if six.PY3 else 'S30'), apertures=apertures, initialize_arrays=True) for i in range(chunk_size)]

        b = ProgressBar(len(sed_files))

        # Loop over SEDs
        for im, sed_file in enumerate(sed_files):

            b.update()

            log.debug('Processing {0}'.format(os.path.basename(sed_file)))

            # Read in SED
            s = SED.read(sed_file, unit_freq=u.Hz, unit_flux=u.mJy, order='nu')

            # Convolve
            for j in range(chunk_size):

                fluxes[j].wavelength = wavelengths[j + jmin]
                fluxes[j].apertures = apertures
                fluxes[j].model_names[im] = s.name

                if n_ap == 1:
                    fluxes[j].flux[im] = s.flux[0, j + jmin]
                    fluxes[j].error[im] = s.error[0, j + jmin]
                else:
                    fluxes[j].flux[im, :] = s.flux[:, j + jmin]
                    fluxes[j].error[im, :] = s.error[:, j + jmin]

        for j in range(chunk_size):
            fluxes[j].sort_to_match(par_table['MODEL_NAME'])
            fluxes[j].write('{0:s}/convolved/MO{1:03d}.fits'.format(model_dir, j + jmin + 1),
                            overwrite=overwrite)
            filters['filter'][j + jmin] = "MO{0:03d}".format(j + jmin + 1)

    return filters
